from typing import Optional, List, Tuple, Dict

import re
import contextlib
from lxml import etree
from os import path
from sphinx.util import logging

from ..typing.hdl import Intf, IntfPort
from ..typing.hdl import Library, LibraryVendor
from ..typing.hdl import Project
from ..directive.string import string_hdl
from .tcl import tcl

logger = logging.getLogger(__name__)


def parse_hdl_regmap(ctime: float, file: str) -> Dict:
    """
    From https://github.com/tfcollins/vger/blob/main/vger/hdl_reg_map.py
    Added methods:
    * USING: import regs and fields from other regmap.
    * WHERE n IS FROM {} to {}: set a repetition pattern for regs and fields.
    """
    regmap = {
        'subregmap': {},
        'owners': [],
        'ctime': ctime
    }

    def get_where(desc: str, reg: str, fi=None) -> Tuple[any]:
        re_expr = r"FROM ([0-9]+) TO ([0-9]+)$"

        m = re.search(re_expr, desc)
        if not bool(m):
            if fi is not None:
                logger.warning(f"Malformed where {desc} in field bits {fi} "
                               f"at reg {reg}!")
            else:
                logger.warning(f"Malformed where {desc} in reg address "
                               f"{reg}!")

            return ''

        if not m.group(1).isdigit() or not m.group(2).isdigit():
            logger.warning(f"Non-numerals in where {desc} in reg address "
                           f"{reg}!")
            return ''

        return (f"Where n is from {m.group(1)} to {m.group(2)}.",
                (int(m.group(1)), int(m.group(2))+1))

    if not path.isfile(file):
        logger.warning(f"{file}: File doesn't exist!")
        return regmap

    with open(file, "r") as f:
        data = f.readlines()
    data = [d.replace("\n", "") for d in data]

    using = []
    while "TITLE" in data:
        # Get title
        tit = data.index("TITLE")
        if data[tit + 1].startswith('USING'):
            while data[tit + 1].startswith('USING'):
                using_ = data[tit + 1][6:].strip()
                tit += 1
                if len(using_) == 0:
                    logger.warning("Malformed using in title entry, skipped!")
                    continue
                using.append(using_)

        title = str(data[tit + 1]).strip()
        title_tool = str(data[tit + 2])
        data = data[tit + 2:]

        if 'ENDTITLE' in [title_tool, title]:
            logger.warning("Malformed title entry, skipped!")
            continue

        regmap['subregmap'][title_tool] = {
            'title': title,
            'using': using,
            'regmap': [],
            'access_type': []
        }

        # Get registers
        access_type = []
        while "REG" in data:
            regi = data.index("REG")
            rfi = data.index("ENDREG")
            reg_params = []

            if not regi:
                break

            reg_where = None
            if data[regi + 1].startswith("0x"):
                reg_import = False

                reg_addr = data[regi + 1]

                reg_d = data[regi + 2]
                where_desc = ''
                if reg_d.startswith("WHERE n IS"):
                    where_desc, reg_where = get_where(reg_d[10:], reg_addr)
                    regi = regi + 1

                reg_name = data[regi + 2].strip()
                reg_desc = [data[f_].replace("''", "``") for f_ in range(regi + 3, rfi)]
                if where_desc != "":
                    reg_desc.append(where_desc)

                try:
                    if '+' in reg_addr:
                        reg_addr = reg_addr.split('+')
                        if reg_addr[1].strip() == 'n':
                            reg_addr_incr = 1
                        else:
                            reg_addr_incr = int(reg_addr[1].replace('*n', ''),
                                                16)
                        reg_addr = int(reg_addr[0], 16)
                        if where_desc == '':
                            logger.warning(f"Ranged addr {reg_addr} without "
                                           f"where method at {reg_name}!")
                    else:
                        reg_addr = int(reg_addr, 16)
                        reg_addr_incr = 0
                        if where_desc != '':
                            logger.warning(f"Static addr {reg_addr} "
                                           f"with where method at {reg_name}!")
                except Exception:
                    logger.warning(f"Malformed register address {reg_addr} "
                                   f"for register {reg_name}.")
                    reg_addr = 0
                    reg_addr_incr = 0
            else:
                reg_import = True
                reg_addr = 0
                reg_addr_incr = 0
                reg_name = data[regi + 1].strip()
                reg_desc = None
                reg_params = []

            with contextlib.suppress(ValueError):
                tet = data.index("TITLE") if "TITLE" in data else -1
                if tet != -1:
                    if regi > tet:
                        # into next regmap
                        break
            data = data[rfi + 1:]

            # Get fields
            fields = []
            while "FIELD" in data:
                fi = data.index("FIELD")
                efi = data.index("ENDFIELD")

                if not fi:
                    break

                if efi < fi:
                        logger.warning(f"Got ENDFIELD without FIELD "
                                       f"for register {reg_name}.")

                with contextlib.suppress(ValueError):
                    if "REG" in data:
                        rege = data.index("REG")
                    else:
                        rege = -1
                    if rege != -1 and fi > rege:
                        # into next register
                        break
                if data[fi + 1].startswith('['):
                    field_where = None
                    field_import = False
                    field_loc = data[fi + 1]
                    field_loc = field_loc.split()
                    field_bits = field_loc[0].replace("[", "").replace("]", "")
                    if field_bits != 'n':
                        delimiters = ["+", "-", "*", "/"]
                        if ':' in field_bits:
                            bits_ = field_bits.split(':')
                        else:
                            bits_ = [field_bits, field_bits]
                        try:
                            bit0_ = int(bits_[0])
                        except Exception:
                            bit0_ = bits_[0]
                            bit_str = bit0_
                            for delimiter in delimiters:
                                bit_str = " ".join(bit_str.split(delimiter))
                            for str_part in bit_str.split():
                                try:
                                    int(str_part)
                                except Exception:
                                    reg_params.append(str_part)
                        try:
                            bit1_ = int(bits_[1])
                        except Exception:
                            bit1_ = bits_[1]
                            bit_str = bit1_
                            for delimiter in delimiters:
                                bit_str = " ".join(bit_str.split(delimiter))
                            for str_part in bit_str.split():
                                try:
                                    int(str_part)
                                except Exception:
                                    reg_params.append(str_part)
                        field_bits = (bit0_, bit1_)

                    if len(field_loc) > 1:
                        field_default = ' '.join(field_loc[1:])
                        field_default_long = ' '.join(field_loc[1:])
                        try:
                            if not field_default.startswith("0x"):
                                raise TypeError("Not hexadecimal")
                            fd_ = int(field_default, 16)
                            if type(field_bits) is tuple:
                                len_f = (field_bits[0] - field_bits[1] + 1)
                                len_d = len(bin(fd_)[2:])
                                if len_d > len_f:
                                    logger.warning("Default value "
                                                   f"'{field_default}' "
                                                   f"overflows field width "
                                                   f"{field_loc[0]} at reg "
                                                   f"'{reg_name}'!")
                            field_default = fd_
                            field_default_long = fd_

                        except Exception:
                            split_field = field_default.split(" = ", 2)
                            if "''" in field_default:
                                logger.warning("Default value "
                                               f"'{field_default}' "
                                               f"contains ''!")
                            field_default = split_field[0].replace("''", "")
                            field_default_long = field_default

                            if "0xX" not in field_default:
                                try:
                                    default_str = split_field[1]
                                    field_default_long = split_field[1]
                                    field_default = f"{split_field[0]}"
                                except Exception:
                                    default_str = split_field[0]
                                default_str = re.sub("`[A-Z0-9_]+", "", default_str)
                                default_str = re.findall("[A-Z0-9_]+", default_str)
                                for str_part in default_str:
                                    try:
                                        int(str_part)
                                    except Exception:
                                        reg_params.append(re.sub('\\[[0-9:]+\\]', ' ', str_part))
                                        # TODO: Match parse_hdl_library extracted parameters
                    else:
                        field_default = None
                        field_default_long = None

                    fi_d = data[fi + 2]
                    where_desc = ''
                    if fi_d.startswith("WHERE n IS"):
                        where_desc, field_where = get_where(fi_d[10:], reg_name,
                                                            field_bits)
                        fi = fi + 1
                        if field_bits != 'n':
                            logger.warning("Where method with field bits "
                                           f"{field_loc[0]} instead of n "
                                           f"at reg '{reg_name}'!")
                    elif field_bits == 'n':
                        logger.warning("No where method for ranged field "
                                       f"n at reg '{reg_name}'!")

                    field_name = data[fi + 2].strip()
                    field_name = field_name.replace("/", "or")
                    field_rw = data[fi + 3]

                    if field_rw == 'R':
                        field_rw = 'RO'
                    elif field_rw == 'W':
                        field_rw = 'WO'
                    if '-V' in field_rw:
                        if 'V' not in access_type:
                            access_type.append('V')
                    field_rw_ = field_rw.replace('-V', '')
                    field_rw = field_rw.replace('-V', 'V')
                    if field_rw_ not in access_type:
                        if field_rw_ not in string_hdl.access_type:
                            logger.warning(f"Malformed access type {field_rw} "
                                           f"for reg {field_name}")
                        else:
                            access_type.append(field_rw)

                    field_desc = [data[f_].replace("''", "``") for f_ in range(fi + 4, efi)]
                    if where_desc != '':
                        field_desc.append(where_desc)
                    if field_default_long is not None and field_default_long != field_default:
                        field_desc.append(f"``{field_default} = {field_default_long}``")

                    fields.append({
                        "import": field_import,
                        "where": field_where,
                        "name": field_name,
                        "bits": field_bits,
                        "default": field_default,
                        "default_long": field_default_long,
                        "rw": field_rw,
                        "description": field_desc,
                    })
                else:
                    for i in range(fi + 1, efi):
                        if any(c in data[i] for c in ('[', ']')) or len(data[i]) == 1:
                            logger.warning(f"Suspicious imported field '{data[i]}' "
                                           f"at imported field group at reg {reg_name}!")
                        fields.append({
                            "import": True,
                            "where": None,
                            "name": data[i].strip(),
                            "bits": None,
                            "default": None,
                            "default_long": None,
                            "rw": None,
                            "description": None,
                        })

                data = data[efi + 1:]

            if len(reg_params):
                reg_params_set = set(reg_params)
                reg_params = list(reg_params_set)
                reg_params.sort()
            regmap['subregmap'][title_tool]['regmap'].append(
                {
                    'import': reg_import,
                    'where': reg_where,
                    'name': reg_name,
                    'address': reg_addr,
                    'addr_incr': reg_addr_incr,
                    'description': reg_desc,
                    'fields': fields,
                    'parameters': reg_params
                }
            )
        regmap['subregmap'][title_tool]['access_type'] = access_type

    return regmap


def resolve_hdl_regmap(rm: Dict) -> None:
    """
    Resolve imported registers and fields at regmaps with the "USING" method.
    parse_hdl_regmap must be called first.
    """

    def patch_field(r, p, r_, p_name):
        for i, j in enumerate(r):
            if j['import']:
                for p_ in p:
                    if j['name'] == p_['name']:
                        r[i] = p_.copy()
                        break
                else:
                    logger.warning(f"Field {j['name']} in reg {p_name} "
                                   f"from import {r_} not found!")

    def patch_reg(r, p, r_):
        def patch_reg_(j, p):
            for p_ in p['regmap']:
                if j['name'] == p_['name']:
                    j['import'] = False
                    j['where'] = p_['where']
                    j['address'] = p_['address']
                    j['addr_incr'] = p_['addr_incr']
                    j['description'] = p_['description']
                    patch_field(j['fields'], p_['fields'], r_, p_['name'])
                    return True
            return False

        for j in r:
            if j['import']:
                # The imported register may have two formats:
                # * implicit, no period (e.g. CTRL):
                #   search all imported register maps
                # * explicit, w/ period (e.g. COMMON.CTRL):
                #   left side is the regmap key, and the right side the reg key
                idx = j['name'].find('.')
                if idx != -1:
                    p_ = j['name'][:idx]
                    j['name'] = j['name'][idx+1:]
                    if p_ in p:
                        patch_reg_(j, p[p_])
                else:
                    for p_ in p:
                        if patch_reg_(j, p[p_]):
                            break
            if j['import']:
                logger.warning(f"Reg {j['name']} in import {r_} not found!")

    def resolve(r):
        using = {}
        for use in r['using']:
            use_ = None
            for i in rm:
                for k in rm[i]['subregmap']:
                    if use == k:
                        use_ = rm[i]['subregmap'][k]
                        break
                else:
                    continue
                break
            if use_ is None:
                logger.warning(f"Couldn't find regmap '{use}'!")
            else:
                using[k] = use_

        patch_reg(r['regmap'], using, r['using'])
        r['using'] = []

    for i in rm:
        for k in rm[i]['subregmap']:
            resolve(rm[i]['subregmap'][k])

    return


def expand_hdl_regmap(rm: Dict) -> None:
    """
    Expand fields with the "WHERE n IS FROM {} TO {}" method.
    resolve_hdl_regmap must be called first.
    This method is not called in the doc generation to avoid clutter.
    Register expansion is done in the SV PKG writer.
    """

    def expand_fields(regmap):
        for r in regmap['regmap']:
            expanded = []
            for f in r['fields']:
                if f['where'] is not None:
                    for n in range(*f['where']):
                        f_ = f.copy()
                        f_['name'] = f_['name'].replace('n', str(n))
                        f_['bits'] = (n, n)
                        expanded.append(f_)
                else:
                    expanded.append(f)
            r['fields'] = expanded

    for i in rm:
        for k in rm[i]['subregmap']:
            expand_fields(rm[i]['subregmap'][k])

    return


def parse_hdl_component(file: str, ctime: float, owners: List = []) -> Dict:
    component = {
        'name': "",
        'bus_interface': {},
        'bus_domain': {},
        'ports': {},
        'parameters': {},
        'ctime': ctime,
        'owners': owners.copy()
    }

    def get_namespaces(item):
        nsmap = item.nsmap
        for i in ['spirit', 'xilinx', 'xsi']:
            if i not in nsmap:
                raise Exception(f"Required namespace {i} not in file!")

        return (nsmap['spirit'], nsmap['xilinx'], nsmap['xsi'])

    def get(item, local_name):
        items = get_all(item, local_name)
        if len(items) == 0:
            return None
        else:
            return items[0]

    def get_all(item, local_name: str):
        template = "/*[local-name()='%s']"
        local_name = local_name.split('/')
        return item.xpath('.' + ''.join([template % ln for ln in local_name]))

    def sattrib(item, attr):
        nonlocal spirit
        return item.get(f"{{{spirit}}}{attr}")

    def xattrib(item, attr):
        nonlocal xilinx
        return item.get(f"{{{xilinx}}}{attr}")

    def stag(item):
        nonlocal spirit
        return item.tag.replace(f"{{{spirit}}}", '')

    def xtag(item):
        nonlocal xilinx
        return item.tag.replace(f"{{{xilinx}}}", '')

    def clean_dependency(string):
        expr = r"spirit:decode\(id\(((.*?))\)\)"
        return re.sub(expr, r"\1", string).replace(')', '').replace('(', '')

    def get_dependency(item, type_=None):
        if type_ is None:
            type_ = stag(item)

        dependency = get(item,
                         f"vendorExtensions/{type_}Info/enablement/isEnabled")
        if dependency is None:
            return None
        else:
            return clean_dependency(xattrib(dependency, 'dependency'))

    def get_range(item):
        min_ = sattrib(item, 'minimum')
        max_ = sattrib(item, 'maximum')
        if max_ is None or min_ is None:
            return None
        else:
            return f"From {min_} to {max_}."

    def get_choice_type(name):
        return name[name.find('_')+1:name.rfind('_')]

    def format_default_value(value, fmt):
        if fmt == "bitString" and value[0:2].lower() == "0x":
            return f"'h{value[2:].upper()}"
        if fmt == "bitString" and value[0] == '"' and value[-1] == '"':
            return f"'b{value[1:-1]}"
        if fmt == "bool":
            return value.title()
        return value

    def format_type_value(string, value):
        if string == 'bitString':
            if value[0:2] == "'h":
                return 'Hex'
            else:
                return 'Bits'
        if string == 'long':
            return 'Integer'
        return string

    def merge_sequential(items):
        """
        Merge generated sequencial signals/buses ending with
        _[qi]?([0-9]+)$ or _([0-9]+)_[pn]$ or _phy([0-9]+)$,
        for example:
          {data_tx_12_p, data_tx_23_p} -> data_tx_*_p
          {data_tx_12, data_tx_23} -> data_tx_*
          {adc_data_i0, adc_data_i0} -> adc_data_i*
          {adc_data_q0, adc_data_q0} -> adc_data_q*
          {rx_phy2, rx_phy4} -> rx_phy*
        However, do not create new IP using _phy*, instead use _*, e.g. phy_*.
        """
        re_expr = r"^([pn]_)?([0-9]+)([iq]|yhp)?(_.*)"  # reversed regex
        for key in list(items):
            m = re.search(re_expr, key[::-1])

            if not bool(m):
                continue

            index = m.group(2)[::-1]

            key_ = re.sub(re_expr, r"\1*\3\4", key[::-1])[::-1]
            if key_ not in items:
                items[key_] = items[key]
                items[key_]['index'] = [index, index]

                re_dep = f"\\s+({index})(\\s+|$|\\))"
                if items[key_]['dependency'] is not None:
                    items[key_]['dependency'] = re.sub(re_dep, " * ", items[key_]['dependency'])

                if 'port_map' in items[key_]:
                    pm = items[key_]['port_map']
                    for k in pm:
                        if pm[k]['dependency'] is not None:
                            m = re.search(re_dep, pm[k]['dependency'])
                            if bool(m):
                                pm[k]['dependency'] = re.sub(re_dep, " * ", pm[k]['dependency'])

                    for inner_key in list(items[key_]['port_map']):
                        inner_key_ = (re.sub(re_expr, r"\1*\3", inner_key[::-1]))[::-1]
                        if (inner_key_ != inner_key):
                            items[key_]['port_map'][inner_key_] = items[key_]['port_map'][inner_key]
                            del items[key_]['port_map'][inner_key]

            else:
                if index < items[key_]['index'][0]:
                    items[key_]['index'][0] = index
                if index > items[key_]['index'][1]:
                    items[key_]['index'][1] = index

            del items[key]

    root = etree.parse(file).getroot()
    spirit, xilinx, _ = get_namespaces(root)
    name = get(root, 'name').text

    component['name'] = name
    bs = component['bus_interface']
    dm = component['bus_domain']
    for bus_interface in get_all(root, 'busInterfaces/busInterface'):
        bus_name = get(bus_interface, 'name').text
        if '_signal_clock' in bus_name:
            signal_name = get(get(bus_interface, 'portMaps/portMap'), 'physicalPort/name').text
            if signal_name not in dm:
                dm[signal_name] = []
            dm[signal_name].append(bus_name[0:bus_name.find('_signal_clock')])
            continue
        if '_signal_reset' in bus_name:
            signal_name = get(get(bus_interface, 'portMaps/portMap'), 'physicalPort/name').text
            if signal_name not in dm:
                dm[signal_name] = []
            dm[signal_name].append(bus_name[0:bus_name.find('_signal_reset')])
            continue

        if get(bus_interface, 'slave') is not None:
            bus_role = 'slave'
        elif get(bus_interface, 'master') is not None:
            bus_role = 'master'
        else:
            bus_role = None

        bs[bus_name] = {
            'name': sattrib(get(bus_interface, 'busType'), 'name'),
            'role': bus_role,
            'dependency': get_dependency(bus_interface, 'busInterface'),
            'port_map': {}
        }

        pm = bs[bus_name]['port_map']
        for port_map in get_all(bus_interface, 'portMaps/portMap'):
            pm[get(port_map, 'physicalPort/name').text] = {
                'logical_port': get(port_map, 'logicalPort/name').text,
                'direction': None,
                'dependency': None,
            }

    lport = component['ports']
    for port in get_all(root, 'model/ports/port'):
        port_name = get(port, 'name').text
        port_direction = get(port, 'wire/direction').text
        port_left = get(port, 'wire/vector/left')
        port_right = get(port, 'wire/vector/right')

        if port_left is not None and port_right is not None:
            port_direction += f" [{port_left.text}:{port_right.text}]"

        dependency = get_dependency(port, 'port')

        found = False
        for bus in bs:
            if port_name in bs[bus]['port_map']:
                found = True
                bs[bus]['port_map'][port_name]['direction'] = port_direction
                bs[bus]['port_map'][port_name]['dependency'] = dependency
                # Can't break because the port_name might be used in more than
                # one bus, e.g. BUS0 enabled if CFG=1 and BUS 2 if CFG=1.

        if found is False:
            lport[port_name] = {
                'direction': port_direction,
                'dependency': dependency,
            }

    merge_sequential(lport)
    merge_sequential(bs)

    pr = component['parameters']
    for parameter in get_all(root, 'parameters/parameter'):
        param_description = get(parameter, 'displayName')
        if param_description is not None:
            param_name = get(parameter, 'name').text
            param_value = get(parameter, 'value')
            param_format = sattrib(param_value, 'format')
            param_default = format_default_value(param_value.text, param_format)
            pr[param_name] = {
                'description': param_description.text,
                'default': param_default,
                'type': format_type_value(param_format, param_default),
                '_choice_ref': sattrib(param_value, 'choiceRef'),
                'choices': None,
                'range': get_range(param_value)
            }

    for parameter in get_all(root, 'model/modelParameters/modelParameter'):
        param_name = get(parameter, 'name').text
        param_type = sattrib(parameter, 'dataType')
        if param_type == "std_logic_vector":
            param_type = "logic vector"
        if param_type is not None and param_name in pr:
            if pr[param_name]['type'] is None:
                pr[param_name]['type'] = param_type.capitalize()
            else:
                param_format = pr[param_name]['type']
                pr[param_name]['type'] = param_format[0].upper()+param_format[1:]

    for choice in get_all(root, 'choices/choice'):
        name = get(choice, 'name').text
        for key in pr:
            if pr[key]['_choice_ref'] == name:
                type_ = get_choice_type(name)
                values = get_all(choice, 'enumeration')
                string = []
                if type_ == 'pairs':
                    for v in values:
                        string.append(f"{sattrib(v, 'text')} ({v.text})")
                elif type_ == 'list':
                    for v in values:
                        string.append(v.text)
                else:
                    break
                pr[key]['choices'] = ', '.join(string)
                break

    return component


def parse_hdl_build_status(file: str) -> Tuple[List, int, List[str]]:

    if not path.isfile(file):
        logger.warning(f"File {file} doesn't exist!")
        return ([], None)

    with open(file, "r") as f:
        data = f.readlines()

    if data[3][0:35] != '| Project | Build number | Status |' \
        or data[4][0:25] != '| --- | --- | --- | --- |' \
        or data[-1][0:11] != '# Finished:':
        logger.warning(f"Malformed build status file {file}.")
        return ([], None)

    try:
        s = 'build number'
        build_number = int(data[0][data[0].find(s)+len(s)+1:])
    except Exception as e:
        build_number = -1
        logger.warning("Couldn't get the build number from the first line "
                       f"of '{file}', exception: {e}.")

    project = []
    for i in range(5, len(data) - 2):
        split = data[i].split('|')
        if len(split) < 4:
            logger.warning(f"Malformed line at {file}:{i}.")
            continue

        project.append([split[1].strip(), 0 if split[3].strip() == "SUCCESS" else 1])

    return (project, build_number)


def parse_hdl_vendor(
    file: str
) -> Tuple[Tuple[str], List[str]]:
    """
    Obtain the carrier from the project vendor file.
    """
    obj = set()
    if not path.isfile(file):
        return ((), ["File doesn't exist!"])

    with open(file, "r") as f:
        data = f.readlines()

    for line in data:
        line = ' '.join(line.split())
        if line.startswith("if [regexp \""):
            m = re.search("if \\[regexp \"_(\\w+)\" \\$project_name", line)
            if not bool(m):
                continue
            obj.add(m.group(1))

    return (tuple(obj), [])


def parse_hdl_library(
    file: str,
) -> Tuple[Optional[LibraryVendor], Optional[str], Optional[str]]:
    """
    Obtain the library dependencies and interfaces from the library file.
    Vendor agnostic, even though we use:
    * adi_ip_files: xilinx
    * ad_ip_file: intel
    Would be better to switch to the same method in the future.
    """
    # library/my_ip/my_ip_monitor/my_ip_monitor_glue_ip.tcl
    # ip_name:   my_ip_monitor_glue (ip inventory (unique per vendor))
    # path_: my_ip/my_ip_monitor (makefile dep (global unique))
    basename = path.basename(file)
    ip_name = basename[:basename.rfind("_")]
    ip_name_ = path.basename(path.dirname(file))
    if ip_name != ip_name_:
        logger.warning(f"{file}: Path basename does not match filename format.")

    idx = file.find('library')
    if idx != -1:
        path_ = path.dirname(file)[idx+8:]
    else:
        logger.warning("'library' not found in path to extract long name'")
        return (None, None, None)

    if not path.isfile(file):
        logger.warning("File doesn't exist!")
        return (None, None, None)

    tcl_ = tcl(file)

    # Check library name against ip_name
    for i, line in enumerate(tcl_):
        for m in ['adi_ip_create', 'ad_ip_create']:
            if line.startswith(m):
                line_ = line.split()
                if ip_name != line_[1]:
                    logger.warning(f"{file}: '{m}' IP name '{line_[1]}' does not "
                                   f"match name '{ip_name}', line {i}")
            break

    # Obtain the file dependencies and top module candidate
    deps = set()
    line = tcl_.line_startswith(["adi_ip_files", "ad_ip_files"])
    if line is None:
        line = tcl_.line_startswith("add_files ")
        top_mod_ = None
    else:
        top_mod_ = line.split()[1]

    if line is not None:
        deps.update(tcl.get_list_items(line))

    for line in tcl_:
        if (line.startswith('add_fileset_file')):
            line_ = line.split()
            if len(line_) >= 5:
                deps.add(line_[4])

    if top_mod_ is None:
        logger.warning(f"{file}: Unable to find top module name for library "
                       f"'{ip_name}', will use the lib name instead.")
        top_mod_ = ip_name

    # Add itself as a dependency
    deps.add(path.basename(file))

    # Obtain the library dependencies
    lib_deps = tcl_.in_method_match("analog\\.com:\\$VIVADO_IP_LIBRARY:(\\w+)",
                                    "adi_ip_add_core_dependencies")

    # Obtain the interface dependencies
    intf = tcl_.in_method_match("analog\\.com:(?:interface|user):(\\w+)",
                                "adi_add_bus")

    # Remove _rtl suffixed
    intf = [e for e in intf if not e.endswith('_rtl')]

    # Find top module name with type
    top_mod = None
    for d in deps:
        if d.startswith(top_mod_) and d[len(top_mod_):] in [".v", ".sv"]:
            top_mod = d
            break

    # Obtain parameters from the top module
    def get_parameters(mod):
        f = path.join(path.dirname(file), mod)
        if not path.isfile(f):
            logger.warning(f"{file}: Top module '{f}' from library '{ip_name}'"
                           " does not exist")
            return None

        with open(f, "r") as f_:
            data = f_.readlines()

        for i, line in enumerate(data):
            if line.strip().startswith("module "):
                break

        if i == len(data) - 1:
            logger.warning(f"{file}: Failed to find 'module' in '{f}'")
            return None

        data = data[i+1:]
        params = []
        for j, line in enumerate(data):
            line = line.strip()
            if line.startswith("parameter "):
                line = re.sub('\\[[0-9:]+:[0-9]+\\]', '', line)
                idx = line.index('=')
                line = [line[:idx], line[idx+1:].strip()]
                if len(line) != 2:
                    logger.warning(f"{file}: Malformed parameter at line {i+j+2} of module '{f}'")
                    continue
                name = line[0].split()[-1]
                value = line[1].replace(',', '')

                params.append((name, value))

            elif line.startswith(')') and line.endswith('('):
                break

        return tuple(params)


    if top_mod is not None:
        param = get_parameters(top_mod)
    else:
        param = None
        logger.warning(f"{file}: Failed to find top module for IP '{ip_name}'")

    # Obtain parameters from the top module
    def get_parameters_ttcl(mod):
        f = path.join(path.dirname(file), mod)
        if not path.isfile(f):
            logger.warning(f"{file}: TTcl '{f}' from library '{ip_name}' does not exist")
            return None

        with open(f, "r") as f_:
            data = f_.readlines()

        param = set()
        for i, line in enumerate(data):
            line = line.split()
            m = "MODELPARAM_VALUE."
            if len(line) == 6 and line[4].startswith(m):
                line = line[4][len(m):-1]
                param.add(line)

        return tuple(param)

    # Find pkg_sv-ttcl, if any
    pkg_sv_ttcl = f"{top_mod_}_pkg_sv.ttcl"
    if pkg_sv_ttcl in deps:
        param_ttcl = get_parameters_ttcl(pkg_sv_ttcl)
    else:
        param_ttcl = None

    if param_ttcl is not None and param is not None:
        # compare
        params_ = [p[0] for p in param]
        for p in param_ttcl:
            if p not in params_:
                logger.warning(f"{file}: Parameter '{p}' in the '{pkg_sv_ttcl}'"
                               f" file not found in the top module '{top_mod}'")
        pass

    obj = LibraryVendor(
        dependencies=tuple(deps),
        library_dependencies=tuple(lib_deps),
        interfaces=tuple(intf),
        parameters=param
    )
    return (obj, path_, ip_name)


def resolve_hdl_library(
    libraries: Dict[str, Library],
    key: str,
    intf_lut: Intf,
) -> None:
    """
    Resolve a library by extracting generic dependencies, resolving paths
    and checking interfaces
    """
    library = libraries[key]

    # Filter generic deps, if:
    # * more than one vendor, the generic files are the intersection
    # * a single vendor, mark as generic if common endings and no vendor name
    deps = {}
    if len(library['vendor']) > 1:
        for v in library['vendor']:
            deps[v] = set(library['vendor'][v]['dependencies'])
        deps['generic'] = set.intersection(*[deps[v] for v in deps])
    else:
        v = list(library['vendor'].keys())[0]
        deps['generic'] = set()
        deps[v] = set()
        for dep in library['vendor'][v]['dependencies']:
            if (dep.endswith(('.v', '.vh', '.vhd', '.sv', 'tcl')) and not
                dep.endswith(('_ip.tcl', '_hw.tcl')) and
                v  not in dep):
                deps['generic'].add(dep)
            else:
                deps[v].add(dep)

    # Expand $ad_hdl_dir and make it relative
    for v in deps:
        for k in deps[v]:
            if k.startswith('$ad_hdl_dir/'):
                lib_path = path.join('library', key)
                relpath = path.relpath(k[12:], lib_path)
                deps[v].add(relpath)
                deps[v].remove(k)

    for v in library['vendor']:
        deps_ = deps[v] - deps['generic']
        library['vendor'][v]['dependencies'] = tuple(sorted(deps_))
    deps_ = tuple(deps['generic'])
    library['generic']['dependencies'] = tuple(sorted(deps_))


    # Find path (relative to hdl/library) of library dependencies
    def resolve_lib_dep(dep):
        for lib in libraries:
            if v in libraries[lib]['vendor']:
                if libraries[lib]['name'] == dep:
                    return lib
        logger.warning(f"Library dependency key '{dep}' not found!")
        return dep
    for v in library['vendor']:
        lib_deps = set()
        for dep in library['vendor'][v]['library_dependencies']:
            lib_deps.add(resolve_lib_dep(dep))
        library['vendor'][v]['library_dependencies'] = tuple(sorted(lib_deps))

    # Find interfaces_ip.tcl source for intf db and obtain
    # Effectively, Xilinx
    for v in library['vendor']:
        deps_intf = set()
        interface_deps = set()
        for intf in library['vendor'][v]['interfaces']:
            if intf not in intf_lut:
                logger.warning(f"Interface {intf} does not exist in any "
                               "interfaces_ip.tcl file.")
            else:
                abs_path = path.join('library', key)
                base = path.relpath(path.join(intf_lut[intf], intf), abs_path)
                deps_intf.add(base + '.xml')
                deps_intf.add(base + "_rtl.xml")
                # XILINX_INTERFACE_DEPS are relative to the library folder
                interface_deps.add(path.relpath(intf_lut[intf], 'library'))
        library['vendor'][v]['interfaces'] = tuple(sorted(deps_intf))
        library['vendor'][v]['interfaces_tcl'] = tuple(sorted(interface_deps))

    return


def parse_hdl_project(
    file: str
) -> Tuple[Optional[Project], Optional[str]]:
    """
    Obtain the project dependencies from the project files.
    Start from system_project.tcl, then walktrough the sourced tcls.
    """

    idx = file.find('projects')
    if idx != -1:
        path_ = path.dirname(file)[idx+9:]
    else:
        logger.warning("'projects' not found in path to extract long name'")
        return (None, None)

    if not path.isfile(file):
        logger.warning(f"{file}: File doesn't exist!")
        return (None, None)

    # Get project name and carrier from path
    project = path.dirname(path_)
    carrier = path.basename(path_)
    if project == "":
        project = carrier
    elif project == "common":
        project = "template"

    base_path = path.dirname(file)
    sys_path = path.join(base_path, "system_project.tcl")
    if not path.isfile(sys_path):
        logger.warning(f"{sys_path}: File doesn't exist!")
        return (None, None)

    tcl_ = tcl(sys_path)
    # Check adi_project, project name and carrier from adi_project*
    project_name = None
    for line in tcl_:
        if line.startswith("adi_project"):
            line_ = line.split()
            project_name = line_[1]
            idx = project_name.find(carrier)
            if idx > 0:
                project_ = line_[1][:idx-1]
                if project_ != project:
                    logger.warning(f"{sys_path}: Project '{project_}' in "
                                   "'adi_project' does not match from path.")
            elif idx == 0:
                pass
            else:
                logger.warning(f"{sys_path}: Carrier from path '{carrier}' "
                                "not found in 'adi_project'.")

            break

    if project_name is None:
        logger.warning(f"{sys_path}: 'adi_project' not found.")
        return (None, None)

    m_deps = set()
    # Get sourced files from system_project.tcl entry point
    tcl_files, tcl_files_ = tcl.get_sourced_files(sys_path)
    m_deps.update(tcl_files_)
    # Get project files
    for t in tcl_files:
        for line in t:
            if line.startswith('adi_project_files'):
                i_ = tcl.get_list_items(line)
                # Convert to relative to file dir
                dir_ = path.dirname(file)
                for f in i_:
                    if f.startswith("$ad_hdl_dir/"):
                        f = f[12:]
                        f = path.relpath(f, dir_)
                    m_deps.add(f)

    # Get sourced files from system_bd/qsys.tcl entry point
    tcl_files, tcl_files_ = tcl.get_sourced_files(file)
    m_deps.update(tcl_files_)
    m_deps = list(m_deps)
    m_deps.sort(reverse=True)
    # Get libraries
    lib_deps = set()
    for t in tcl_files:
        for line in t:
            if line.startswith('ad_ip_instance'):
                l_ = line.split()
                if len(l_) > 1:
                    lib_deps.add(l_[1])
            elif line.startswith('add_instance'):
                l_ = line.split()
                if len(l_) > 2:
                    lib_deps.add(l_[2])


    obj = Project(
        name=project_name,
        lib_deps=tuple(lib_deps),
        m_deps=tuple(m_deps)
    )
    return (obj, path_)


def resolve_hdl_project(
    project: Project,
    libraries: Dict[str, Library],
) -> None:
    def find_lib(key, vendor):
        """
        Find path (relative to hdl/library) of library dependencies
        If not found, consider as third_party and don't include to set
        """
        for lib in libraries:
            if key in lib and vendor in libraries[lib]['vendor']:
                return lib
        return

    lib_deps = set()
    for lib in project['lib_deps']:
        if lib_ := find_lib(lib, project['vendor']):
            lib_deps.add(lib_)
    lib_deps = list(lib_deps)
    lib_deps.sort()

    project['lib_deps'] = tuple(lib_deps)
    return


def parse_hdl_interfaces(
    file: str,
) -> Tuple[Intf]:
    """
    Obtain the interfaces from the interfaces file.
    """
    obj = []

    if not path.isfile(file):
        logger.warning(f"{file}: File doesn't exist!")
        return ()

    with open(file, "r") as f:
        data = f.readlines()

    intf = None
    descr = None
    for i, line in enumerate(data):
        line = line.strip()
        if len(line) > 0 and line[0] == '#':
            descr = line[1:].strip()
        if line.startswith('adi_if_define'):
            intf = line.split()[1].replace('"', '')
            obj.append(Intf(
                description=descr,
                name=intf,
                ports=[]
            ))
            descr = None

        if line.startswith('adi_if_ports'):
            if len(obj) == 0:
                logger.warning(f"{file}: 'adi_if_ports' at line {i+1} "
                                "without precending adi_if_ports")
                continue
            ports = line.split()
            try:
                if len(ports) < 4:
                    raise Exception(f"too few arguments, got {len(ports)}")
                if ports[1] not in ['input', 'output']:
                    raise Exception(f"unknown direction '{ports[1]}'")
                direction = ports[3]
                width = int(ports[2])
                name = ports[3]
                domain = 'none' if len(ports) < 5 else ports[4]
                domain = None if domain == 'none' else domain
                default = None if len(ports) < 6 else int(ports[5])

                obj[-1]['ports'].append(IntfPort(
                    direction=direction,
                    width=width,
                    name=name,
                    domain=domain,
                    default=default,
                ))
            except Exception as e:
                logger.warning(f"{file}: Malformed 'adi_if_ports' at line {i+1}, "
                               f"exception: {e}")
    for o in obj:
        o['ports'] = tuple(o['ports'])
    return tuple(obj)



from lxml import etree

font_size = 16
margin = 18
line_length = 16
text_vertical_margin = 8
color_main = '#0067b9'
color_bg1 = '#c4e5ff'
color_bg2 = '#ebf6ff'
stroke_width = 3


class hdl_component():
    @staticmethod
    def get_name(lib_name):
        return f"{lib_name.replace('/','-')}.svg"

    @staticmethod
    def render(lib_name, item):
        def make_style(parent):
            etree.SubElement(parent, 'style').text = """
                a {
                    text-decoration: none;
                }
                """

        def make_gradient(parent):
            defs = etree.SubElement(parent, 'defs')
            gradient = etree.SubElement(defs, 'linearGradient', attrib={
                'id': 'ip_background',
                'x1': '0', 'x2': '1', 'y1': '0', 'y2': '1'
                })
            etree.SubElement(gradient, 'stop', attrib={'offset': '0%', 'stop-color': color_bg1})
            etree.SubElement(gradient, 'stop', attrib={'offset': '100%', 'stop-color': color_bg2})

        def symbol_bus(parent):
            rect_height = font_size
            one_fifth_x = line_length/5
            one_fifth_y = rect_height/5
            x = 0
            y = 0
            pattern = [[0, 1, 0, 1, 0], [0, 0, 0, 0, 0], [0, 1, 0, 1, 0],
                       [0, 0, 0, 0, 0], [0, 1, 0, 1, 0]]
            etree.SubElement(parent, "rect", attrib={
                'x': '0', 'y': '0',
                'width': str(line_length), 'height': str(font_size),
                'fill': color_bg1,
            })
            for i in pattern:
                for j in i:
                    if j:
                        etree.SubElement(parent, "rect", attrib={
                            'x': str(x), 'y': str(y),
                            'width': str(one_fifth_x), 'height': str(one_fifth_y),
                            'fill': color_main,
                        })
                    x = x+one_fifth_x
                x = 0
                y = y+one_fifth_y

        def symbol_port(parent):
            etree.SubElement(parent, "line", attrib={
                'stroke': 'white',
                'stroke-width': str(stroke_width*2),
                'x1': '0', 'y1': str(font_size/2),
                'x2': str(line_length+stroke_width), 'y2': str(font_size/2)
                })
            etree.SubElement(parent, "line", attrib={
                'stroke': 'black',
                'stroke-width': str(stroke_width),
                'x1': '0', 'y1': str(font_size/2),
                'x2': str(line_length), 'y2': str(font_size/2)
                })

        def create_text(items, side):
            y_pos = margin*3
            if side.startswith('out'):
                text_anchor = 'end'
                x_pos = margin*3 + aux_width
                x_pos_group = x_pos+margin
                scale_group = 'scale(1,1)'
            else:
                text_anchor = 'start'
                x_pos = margin*2
                x_pos_group = x_pos-margin
                scale_group = 'scale(-1,1)'

            for elem in items:
                if elem[1] == 'bus':
                    link_anchor = f"#bus-interface-{elem[0]}"
                else:
                    link_anchor = "#ports"
                link = etree.SubElement(root, "a", attrib={
                    'href': link_anchor,
                    })
                etree.SubElement(link, "text", attrib={
                    'style': f"font: {font_size}px sans-serif",
                    'text-anchor': text_anchor,
                    'dominant-baseline': 'middle',
                    'x': str(x_pos), 'y': str(y_pos)
                    }).text = elem[0]
                group = etree.SubElement(root, "g", attrib={
                    'transform': f"translate({x_pos_group},{y_pos-font_size/2}) {scale_group}"}
                    )
                if elem[1] == 'bus':
                    symbol_bus(group)
                else:
                    symbol_port(group)
                y_pos += font_size+text_vertical_margin

        ins = []
        outs = []
        for key in item['bus_interface']:
            if item['bus_interface'][key]['role'] == 'master':
                outs.append((key, 'bus'))
            else:
                ins.append((key, 'bus'))
        for key in item['ports']:
            if item['ports'][key]['direction'].startswith('out'):
                outs.append((key, 'port'))
            else:
                ins.append((key, 'port'))

        max_len_in = 0
        for elem in ins:
            if len(elem[0]) > max_len_in:
                max_len_in = len(elem[0])

        max_len_out = 0
        for elem in outs:
            if len(elem[0]) > max_len_out:
                max_len_out = len(elem[0])

        aux_width = (max_len_in+max_len_out)*font_size*.6

        num_outs = len(outs)
        num_ins = len(ins)
        max_num = max(num_outs, num_ins)

        root = etree.Element('svg', xmlns="http://www.w3.org/2000/svg")

        make_style(root)
        make_gradient(root)

        ip_width = aux_width + margin*3
        ip_height = max_num*(font_size+text_vertical_margin) + margin*2
        etree.SubElement(root, "rect", attrib={
            'x': str(margin), 'y': str(margin),
            'width': str(ip_width), 'height': str(ip_height),
            'rx': str(margin),
            'fill': 'url(#ip_background)'
        })

        create_text(ins, 'in')
        create_text(outs, 'out')

        viewbox_x = margin*5 + aux_width
        viewbox_y = margin*5 + max_num*(font_size+text_vertical_margin)
        root.set('viewBox', f"0 0 {viewbox_x} {viewbox_y}")
        root.set('width', str(viewbox_x))
        root.set('class', 'no-background')

        etree.SubElement(root, "rect", attrib={
            'x': str(margin), 'y': str(margin),
            'width': str(ip_width), 'height': str(ip_height),
            'rx': str(margin),
            'fill': 'none',
            'stroke': color_main,
            'stroke-width': str(stroke_width)
        })

        ipname_y = viewbox_y-font_size
        ipname_x = viewbox_x/2
        etree.SubElement(root, "text", attrib={
            'style': f"font: {font_size}px sans-serif",
            'fill': color_main,
            'text-anchor': 'middle',
            'dominant-baseline': 'middle',
            'x': str(ipname_x), 'y': str(ipname_y)
            }).text = lib_name[lib_name.rfind('/')+1:]

        return etree.ElementTree(root)

    @staticmethod
    def render_placeholder(lib_name):
        root = etree.Element('svg', xmlns="http://www.w3.org/2000/svg")

        def text_element(text, fs, x, y, font='sans-serif'):
            etree.SubElement(root, "text", attrib={
                'style': f"font: {font_size*fs}px {font}",
                'fill': '#666',
                'text-anchor': 'middle',
                'dominant-baseline': 'middle',
                'x': str(x), 'y': str(y)
                }).text = text

        ip_width = font_size*30
        ip_height = font_size*15
        etree.SubElement(root, "rect", attrib={
            'x': str(margin), 'y': str(margin),
            'width': str(ip_width), 'height': str(ip_height),
            'rx': str(margin),
            'stroke': '#666',
            'fill': 'none',
            'stroke-width': str(stroke_width),
            'stroke-dasharray': '8 16',
            'stroke-linecap': 'round'
        })

        viewbox_x = ip_width + margin*2
        viewbox_y = ip_height + margin*2
        root.set('viewBox', f"0 0 {viewbox_x} {viewbox_y}")
        root.set('width', str(viewbox_x))
        root.set('class', 'no-background')

        text_y = viewbox_y/2.5
        text_x = viewbox_x/2
        text_element("?", 4, text_x, text_y-text_vertical_margin*2)
        text_element(f"{lib_name[lib_name.rfind('/')+1:]} IP-XACT not found.",
                     1, text_x, text_y+font_size*2+text_vertical_margin)
        text_element("Generate it and the documentation:",
                     .75, text_x, text_y+font_size*3+text_vertical_margin*2.5)
        text_element(f"(cd {lib_name}; make)",
                     .75, text_x, text_y+font_size*4+text_vertical_margin*3,
                     'monospace')
        text_element("(cd docs; make html)",
                     .75, text_x, text_y+font_size*5+text_vertical_margin*3.5,
                     'monospace')

        return etree.ElementTree(root)



"""Search searchindex.js and relate to objects.inv
Translated to python from
./adi_doctools/themes/harmonic/scripts/search.js
That is a fork from sphinx provided search.
"""

import asyncio
import hashlib
import json
import os
import pickle
import re
import shutil
import sys
import time
from packaging.version import Version
from concurrent.futures import ThreadPoolExecutor
from html.parser import HTMLParser
from io import StringIO
from urllib.request import urlopen
from urllib.error import URLError, HTTPError
from urllib.parse import urljoin, urlparse

import click
import snowballstemmer

from ..lut import repos, remote_doc, source_hostname_raw
from .aux_html2md import convert_html_to_markdown

from pathlib import Path
from sphinx import __version__ as __sphinx_version__
from sphinx.ext.intersphinx._load import _InvConfig

if Version(__sphinx_version__) < Version('9.0.0'):
    from sphinx.ext.intersphinx._load import _fetch_inventory
else:
    from sphinx.ext.intersphinx._load import _fetch_inventory_data, _load_inventory


BLUE = '\033[94m'
RESET = '\033[0m'

CACHE_DIR = Path('/tmp/adoc.search')
CACHE_VALIDITY = 3600  # 1 hour
RESULTS_FILE = CACHE_DIR / f'last_results_{hashlib.md5(os.getcwd().encode()).hexdigest()}.json'

STOPWORDS = {
    'a', 'and', 'are', 'as', 'at',
    'be', 'but', 'by',
    'for',
    'if', 'in', 'into', 'is', 'it',
    'near', 'no', 'not',
    'of', 'on', 'or',
    'such',
    'that', 'the', 'their', 'then', 'there', 'these', 'they', 'this', 'to',
    'was', 'will', 'with',
}

SCORE_TITLE = 15
SCORE_PARTIAL_TITLE = 8
SCORE_TERM = 5
SCORE_PARTIAL_TERM = 2


def split_strings_in_tuple(t):
    out = []
    for item in t:
        if isinstance(item, str):
            out.extend(item.split())
        else:
            out.append(item)
    return tuple(out)


def get_terminal_width():
    """Get the current terminal width."""
    return shutil.get_terminal_size().columns


def get_terminal_height():
    """Get the current terminal height."""
    return shutil.get_terminal_size().lines


def calculate_limit_from_terminal():
    """Calculate result limit based on terminal height.

    The result takes 6 lines
    - Title line
    - URL line
    - References line (label refs + doc ref)
    - Summary line 1
    - Summary line 2
    - Blank line

    10 is the header size.
    """
    height = get_terminal_height()
    limit = max(3, (height - 10) // 6)
    return limit


def calculate_wrapped_lines(text, width):
    """Calculate how many lines text will take when wrapped."""
    if not text:
        return 1

    # Strip ANSI codes
    ansi_escape = re.compile(r'\x1b\[[0-9;]*m|\x1b\]8;;[^\x1b]*\x1b\\|\x1b\]8;;\x1b\\')
    visible_text = ansi_escape.sub('', text)
    visible_length = len(visible_text)

    return max(1, (visible_length + width - 1) // width)


def truncate_text(text, max_chars, query_terms=None, add_ellipsis=True):
    """Truncate text to fit within max characters."""
    # Strip all ANSI codes
    ansi_escape = re.compile(r'\x1b\[[0-9;]*m|\x1b\]8;;[^\x1b]*\x1b\\|\x1b\]8;;\x1b\\')
    plain_text = ansi_escape.sub('', text)

    if len(plain_text) <= max_chars:
        result = plain_text
    else:
        if add_ellipsis:
            result = plain_text[:max_chars - 3] + "..."
        else:
            result = plain_text[:max_chars]

    if query_terms:
        for term in query_terms:
            if term.lower() not in STOPWORDS:
                pattern = re.compile(re.escape(term), re.IGNORECASE)
                result = pattern.sub(lambda m: click.style(m.group(), fg='yellow'), result)

    return result


def move_cursor_up(lines):
    """Move cursor up by specified number of lines."""
    if lines > 0 and sys.stdout.isatty():
        sys.stdout.write(f'\033[{lines}A')
        sys.stdout.flush()


def move_cursor_down(lines):
    """Move cursor down by specified number of lines."""
    if lines > 0 and sys.stdout.isatty():
        sys.stdout.write(f'\033[{lines}B')
        sys.stdout.flush()


def clear_line():
    """Clear the current line."""
    if sys.stdout.isatty():
        sys.stdout.write('\r\033[2K')
        sys.stdout.flush()


def make_clickable_link(url, display_text=None):
    """Create a clickable terminal link using OSC 8."""
    if display_text is None:
        display_text = url

    if not sys.stdout.isatty():
        # In non-TTY mode, just return the display text (no hyperlink codes)
        return display_text

    # OSC 8 format: \033]8;;URL\033\\DISPLAY_TEXT\033]8;;\033\\
    return f"\033]8;;{url}\033\\{display_text}\033]8;;\033\\"


def extract_path_from_url(url):
    """Extract a pathname from URL for display."""
    if '://' in url:
        path = url.split('://', 1)[1]
    else:
        path = url

    if '/' in path:
        parts = path.split('/', 1)
        if len(parts) > 1:
            path = parts[1]

    if '#' in path:
        path = path.split('#')[0]

    if path.endswith('.html'):
        path = path[:-5]

    return path



class HTMLTextExtractor(HTMLParser):
    """Extract text content from HTML, skipping scripts and styles."""

    def __init__(self):
        super().__init__()
        self.text = StringIO()
        self.skip = False

    def handle_starttag(self, tag, attrs):
        if tag in ('script', 'style'):
            self.skip = True

    def handle_endtag(self, tag):
        if tag in ('script', 'style'):
            self.skip = False

    def handle_data(self, data):
        if not self.skip:
            self.text.write(data)

    def get_text(self):
        return self.text.getvalue()


def get_cache_path(url):
    """Get the cache file path for a given URL."""
    parsed = urlparse(url)

    parts = [parsed.netloc]

    path = parsed.path.strip('/')
    path = path[:-len("/searchindex.js")]
    parts.append(path.replace('/', '.'))

    filename = '.'.join(parts) + '.json'

    return CACHE_DIR / filename


def is_cache_valid(cache_path):
    """Check if cache file exists and is less than 10 minutes old."""
    if not cache_path.exists():
        return False

    file_age = time.time() - cache_path.stat().st_mtime
    return file_age < CACHE_VALIDITY


def load_from_cache(cache_path):
    """Load search index from cache file."""
    with open(cache_path, 'r', encoding='utf-8') as f:
        data = json.load(f)

    if isinstance(data, dict) and data.get('__cache_error__'):
        return None

    return data


def save_to_cache(cache_path, data):
    """Save search index to cache file."""
    # Ensure cache directory exists
    cache_path.parent.mkdir(parents=True, exist_ok=True)

    with open(cache_path, 'w', encoding='utf-8') as f:
        json.dump(data, f)


def save_error_to_cache(cache_path, error_type, url):
    """Save an error marker to cache file."""
    cache_path.parent.mkdir(parents=True, exist_ok=True)

    error_marker = {
        '__cache_error__': True,
        'error_type': error_type,
        'url': url,
        'timestamp': time.time()
    }

    with open(cache_path, 'w', encoding='utf-8') as f:
        json.dump(error_marker, f)


def save_search_results(results):
    """Save search results to temp file for --fetch option."""
    RESULTS_FILE.parent.mkdir(parents=True, exist_ok=True)

    results_data = []
    for i, (title, url, score, repo_name, doc_ref, label_refs, docname) in enumerate(results, 1):
        results_data.append({
            'index': i,
            'title': title,
            'url': url,
            'score': score,
            'repo': repo_name,
            'docname': docname,
            'doc_ref': doc_ref
        })

    with open(RESULTS_FILE, 'w', encoding='utf-8') as f:
        json.dump(results_data, f, indent=2)


def load_search_results():
    """Load previous search results from temp file."""
    if not RESULTS_FILE.exists():
        return None

    try:
        with open(RESULTS_FILE, 'r', encoding='utf-8') as f:
            return json.load(f)
    except Exception:
        return None


def fetch_url_content(url, format='md'):
    """Fetch content from URL in specified format."""
    if format == 'src':
        click.echo("Error: --format src is not applicable for URL fetch.", err=True)
        click.echo("The 'src' format fetches source files (.rst/.md) from repositories,", err=True)
        click.echo("which requires using an index from previous search results.", err=True)
        raise click.Abort()

    click.echo("")
    click.echo(f"{BLUE}Format:{RESET} {format.upper()}", nl=False)
    if format == 'md':
        click.echo(" (converted to markdown)")
    elif format == 'html':
        click.echo(" (html)")

    click.echo(click.style("Available: md (default), html, src", dim=True))
    click.echo(f"\n{BLUE}Fetching:{RESET} {url}\n")

    try:
        with urlopen(url, timeout=30) as response:
            html_content = response.read().decode('utf-8')
    except HTTPError as e:
        click.echo(f"Error: HTTP {e.code} - {url}", err=True)
        raise click.Abort()
    except URLError as e:
        click.echo(f"Error: Failed to fetch URL - {e.reason}", err=True)
        raise click.Abort()
    except Exception as e:
        click.echo(f"Error: {e}", err=True)
        raise click.Abort()

    click.echo("─" * get_terminal_width())

    if format == 'html':
        click.echo(html_content)
    elif format == 'md':
        markdown = convert_html_to_markdown(url, html_content)
        if markdown is None:
            click.echo("Error: Failed to convert HTML to Markdown", err=True)
            raise click.Abort()
        click.echo(markdown)

def fetch_source_file(index, format='src'):
    """Fetch and display file from previous search results in specified format."""
    results = load_search_results()

    if results is None:
        click.echo("Error: No previous search results found. Run a search first.", err=True)
        raise click.Abort()

    result = None
    for r in results:
        if r['index'] == index:
            result = r
            break

    if result is None:
        click.echo(f"Error: No result found at index {index}. Valid indices: 1-{len(results)}", err=True)
        raise click.Abort()

    result_url = result['url']

    click.echo("")
    click.echo(f"{BLUE}Format:{RESET} {format.upper()}", nl=False)
    if format == 'md':
        click.echo(" (converted to markdown)")
    elif format == 'src':
        click.echo(" (source reST or markdown)")
    elif format == 'html':
        click.echo(" (html)")
    click.echo(click.style("Available: md (default), src, html", dim=True))

    if format == 'html':
        click.echo(f"\n{BLUE}Fetching from:{RESET} [{index}] {result['title']}")
        click.echo(f"{BLUE}URL:{RESET} {result_url}\n")

        try:
            with urlopen(result_url, timeout=30) as response:
                html_content = response.read().decode('utf-8')
        except Exception as e:
            click.echo(f"Error: Failed to fetch HTML - {e}", err=True)
            raise click.Abort()

        click.echo("─" * get_terminal_width())
        click.echo(html_content)
        return

    elif format == 'md':
        click.echo(f"\n{BLUE}Fetching from:{RESET} [{index}] {result['title']}")
        click.echo(f"{BLUE}URL:{RESET} {result_url}\n")

        try:
            with urlopen(result_url, timeout=30) as response:
                html_content = response.read().decode('utf-8')
        except Exception as e:
            click.echo(f"Error: Failed to fetch HTML - {e}", err=True)
            raise click.Abort()

        markdown = convert_html_to_markdown(result_url, html_content)
        if markdown is None:
            click.echo("Error: Failed to convert HTML to Markdown", err=True)
            raise click.Abort()

        click.echo("─" * get_terminal_width())
        click.echo(markdown)
        return

    # format == 'src' - fetch source file
    repo_name = result['repo']
    docname = result['docname']

    if not repo_name or repo_name not in repos:
        click.echo(f"Error: Unknown source location for '{repo_name or 'unknown'}'.", err=True)
        click.echo("  " + ", ".join(sorted(repos.keys())), err=True)
        raise click.Abort()

    repo_info = repos[repo_name]
    pathname = repo_info['pathname']
    branch = repo_info['branch']

    # Construct the source file path
    # docname is like "library/axi_dmac/index" -> "library/axi_dmac/index.rst"
    # or could be other extensions
    source_extensions = ['.rst', '.md']

    click.echo(f"\n{BLUE}Fetching from:{RESET} [{index}] {result['title']}")
    click.echo(f"{BLUE}Repository:{RESET} {repo_name}")
    click.echo(f"{BLUE}Document:{RESET} {docname}\n")

    source_url = None
    content = None

    for ext in source_extensions:
        # Build the full path: pathname/docname.ext
        if pathname:
            full_path = f"{pathname}/{docname}{ext}"
        else:
            full_path = f"{docname}{ext}"

        # Use source_hostname_raw template
        test_url = source_hostname_raw.format(
            repository=repo_name,
            branch=branch,
            pathname=full_path
        )

        try:
            response = urlopen(test_url)
            content = response.read().decode('utf-8')
            source_url = test_url
            break
        except HTTPError as e:
            if e.code == 404:
                continue
            else:
                raise
        except Exception:
            continue

    if content is None:
        click.echo(f"Error: Could not fetch source file. Tried extensions: {', '.join(source_extensions)}", err=True)
        click.echo(f"URL: {result['url']}", err=True)
        raise click.Abort()

    include_pattern = re.compile(r'^\s*\.\.\s+include::\s+(.+?)\s*$', re.MULTILINE)
    lines = [line for line in content.strip().split('\n') if line.strip() and not line.strip().startswith('..') or line.strip().startswith('.. include::')]

    if len(lines) <= 1:
        match = include_pattern.search(content)
        if match:
            include_path = match.group(1).strip()

            # Resolve the include path relative to the current file
            # The current file path is like "doc/sphinx/source/drivers/adc/ad405x.rst"
            # The include path is like "../../../../../drivers/adc/ad405x/README.rst"

            if pathname:
                current_file_path = f"{pathname}/{docname}"
            else:
                current_file_path = docname

            current_dir = '/'.join(current_file_path.split('/')[:-1]) if '/' in current_file_path else ''

            if current_dir:
                resolved_path = f"{current_dir}/{include_path}"
            else:
                resolved_path = include_path

            path_parts = []
            for part in resolved_path.split('/'):
                if part == '..':
                    if path_parts:
                        path_parts.pop()
                elif part and part != '.':
                    path_parts.append(part)

            resolved_path = '/'.join(path_parts)

            included_url = source_hostname_raw.format(
                repository=repo_name,
                branch=branch,
                pathname=resolved_path
            )

            try:
                click.echo(f"{BLUE}Following include directive to:{RESET} {include_path}\n")
                response = urlopen(included_url)
                content = response.read().decode('utf-8')
                source_url = included_url
            except HTTPError as e:
                if e.code == 404:
                    click.echo("Warning: Could not fetch included file (404). Showing original content.", err=True)
                    click.echo(f"Attempted URL: {included_url}\n", err=True)
                else:
                    click.echo(f"Warning: Could not fetch included file (HTTP {e.code}). Showing original content.", err=True)
            except Exception as e:
                click.echo(f"Warning: Could not fetch included file: {e}. Showing original content.", err=True)

    click.echo(f"{BLUE}Source URL:{RESET} {source_url}\n")
    click.echo("─" * get_terminal_width())
    click.echo(content)


def get_inventory_cache_path(base_url):
    """Get cache path for objects.inv based on base URL."""
    inv_url = base_url.rstrip('/') + '/objects.inv'
    parsed = urlparse(inv_url)

    parts = [parsed.netloc]

    path = parsed.path.strip('/')
    path = path[:-len("/objects.inv")]
    parts.append(path.replace('/', '.'))

    filename = '.'.join(parts) + '.inv.pickle'
    return CACHE_DIR / filename


def load_inventory_from_cache(cache_path):
    """Load inventory from cache file."""
    with open(cache_path, 'rb') as f:
        # First byte indicates if it's an error marker
        first_bytes = f.read(10)
        f.seek(0)

        if first_bytes.startswith(b'{'):
            f.close()
            with open(cache_path, 'r', encoding='utf-8') as jf:
                data = json.load(jf)
                if isinstance(data, dict) and data.get('__cache_error__'):
                    return None

        return pickle.load(f)


def save_inventory_to_cache(cache_path, inventory):
    """Save inventory to cache file."""
    cache_path.parent.mkdir(parents=True, exist_ok=True)

    with open(cache_path, 'wb') as f:
        pickle.dump(inventory, f)


def save_inventory_error_to_cache(cache_path, error_type, url):
    """Save an error marker for inventory cache."""
    cache_path.parent.mkdir(parents=True, exist_ok=True)

    error_marker = {
        '__cache_error__': True,
        'error_type': error_type,
        'url': url,
        'timestamp': time.time()
    }

    with open(cache_path, 'w', encoding='utf-8') as f:
        json.dump(error_marker, f)


def fetch_search_index(url):
    """Fetch and parse searchindex.js from URL, with caching."""
    cache_path = get_cache_path(url)

    if is_cache_valid(cache_path):
        cached_data = load_from_cache(cache_path)
        if cached_data is None:
            return None
        return cached_data

    try:
        with urlopen(url, timeout=30) as response:
            content = response.read().decode('utf-8')
    except HTTPError as e:
        if e.code == 404:
            save_error_to_cache(cache_path, '404', url)
            return None
        raise URLError(f"HTTP {e.code}: {url}")
    except URLError as e:
        raise URLError(f"Failed to fetch {url}: {e.reason}")

    match = re.search(r'Search\.setIndex\((.*)\)', content, re.DOTALL)
    if not match:
        raise ValueError("Could not find Search.setIndex() in searchindex.js")

    json_str = match.group(1)
    try:
        data = json.loads(json_str)
        save_to_cache(cache_path, data)
        return data
    except json.JSONDecodeError as e:
        raise ValueError(f"Failed to parse searchindex.js JSON: {e}")


def fetch_intersphinx_inventory(base_url):
    """Fetch and parse intersphinx objects.inv file, with caching."""
    if not base_url.endswith('/'):
        base_url += '/'

    inv_url = base_url + 'objects.inv'
    cache_path = get_inventory_cache_path(base_url)

    if is_cache_valid(cache_path):
        try:
            cached_inv = load_inventory_from_cache(cache_path)
            if cached_inv is None:
                return None
            return cached_inv
        except Exception:
            pass

    config = _InvConfig(
        intersphinx_cache_limit=5,
        intersphinx_timeout=10,
        tls_verify=False,
        tls_cacerts=None,
        user_agent='',
    )

    try:
        if Version(__sphinx_version__) < Version('9.0.0'):
            inv = _fetch_inventory(
                target_uri='',
                inv_location=inv_url,
                config=config,
                srcdir=Path(),
            )
        else:
            raw_data, target_uri = _fetch_inventory_data(
                target_uri='',
                inv_location=inv_url,
                config=config,
                srcdir=Path(),
                cache_path=None,
            )
            inv = _load_inventory(raw_data, target_uri=target_uri)
        save_inventory_to_cache(cache_path, inv)
        return inv
    except HTTPError as e:
        if e.code == 404:
            save_inventory_error_to_cache(cache_path, '404', inv_url)
            return None
        raise
    except Exception:
        return None


def get_intersphinx_references(inventory, docname, repo_name, base_url):
    """Get intersphinx references for a document."""
    if not inventory:
        return {'doc_ref': None, 'label_refs': []}

    result = {'doc_ref': None, 'label_refs': []}

    if 'std:doc' in inventory.data and docname in inventory.data['std:doc']:
        if repo_name:
            result['doc_ref'] = f":external+{repo_name}:doc:`{docname}`"
        else:
            result['doc_ref'] = f":std:doc:`{docname}`"

    if 'std:label' in inventory.data:
        doc_uri_prefix = f"{docname}.html#"

        for label_name, inv_item in inventory.data['std:label'].items():
            if inv_item.uri.startswith(doc_uri_prefix):
                full_url = urljoin(base_url, inv_item.uri)

                if repo_name:
                    ref_text = f":external+{repo_name}:ref:`{label_name}`"
                else:
                    ref_text = f":std:label:`{label_name}`"

                result['label_refs'].append((ref_text, label_name, full_url))

    return result


def get_base_url(searchindex_url):
    """Extract base documentation URL from searchindex.js URL."""
    return searchindex_url.rsplit('/', 1)[0] + '/'


def stem_query(query_terms, stemmer):
    """Stem query terms and filter stopwords."""
    stemmed = []
    for term in query_terms:
        term_lower = term.lower()
        if term_lower not in STOPWORDS:
            stemmed.append(stemmer.stemWord(term_lower))
    return stemmed


def extract_html_text(html_content, anchor=None):
    """Extract text from HTML content."""
    try:
        # Try to find [role="main"] section
        main_match = re.search(r'<[^>]+role=["\']main["\'][^>]*>(.*)',
                              html_content, re.DOTALL | re.IGNORECASE)

        if main_match:
            content = main_match.group(1)
        else:
            content = html_content

        # If anchor specified, try to extract that section
        if anchor:
            anchor_id = anchor.lstrip('#')
            patterns = [
                rf'<[^>]+id=["\']?{re.escape(anchor_id)}["\']?[^>]*>(.*?)(?=<h[1-6][^>]*>|</section>|</div>|$)',
                rf'id=["\']?{re.escape(anchor_id)}["\']?[^>]*>([^<]*.*?)(?=<h[1-6]|</(?:div|section)|$)',
            ]

            for pattern in patterns:
                anchor_match = re.search(pattern, content, re.DOTALL | re.IGNORECASE)
                if anchor_match:
                    content = anchor_match.group(1)
                    break

        extractor = HTMLTextExtractor()
        extractor.feed(content)
        text = extractor.get_text()

        text = re.sub(r'\s+', ' ', text).strip()
        return text
    except Exception:
        return ""


def get_summary(html_content, query_terms, anchor=None, max_chars=480):
    """Extract a summary from HTML content around the first matching keyword."""
    text = extract_html_text(html_content, anchor)
    if not text:
        return None

    text_lower = text.lower()

    positions = []
    for term in query_terms:
        term_lower = term.lower()
        pos = text_lower.find(term_lower)
        if pos >= 0:
            positions.append(pos)

    if not positions:
        start_pos = 0
    else:
        start_pos = max(min(positions) - 180, 0)

    summary = text[start_pos:start_pos + max_chars].strip()

    for term in query_terms:
        if term.lower() in STOPWORDS:
            continue
        pattern = re.compile(re.escape(term), re.IGNORECASE)
        summary = pattern.sub(lambda m: click.style(m.group(), fg='yellow'), summary)

    prefix = "..." if start_pos > 0 else ""
    suffix = "..." if start_pos + max_chars < len(text) else ""

    summary = prefix + summary + suffix

    return summary


def search_index(index_data, query_terms, stemmer):
    """Search the index for query terms."""
    results = {}

    docnames = index_data.get('docnames', [])
    titles = index_data.get('titles', [])
    alltitles = index_data.get('alltitles', {})
    titleterms = index_data.get('titleterms', {})
    terms = index_data.get('terms', {})

    # Stem query terms
    stemmed_terms = stem_query(query_terms, stemmer)
    if not stemmed_terms:
        return results

    query_lower = ' '.join(term.lower() for term in query_terms)

    # 1. Search in alltitles (exact and partial title matches)
    for title_text, occurrences in alltitles.items():
        title_lower = title_text.lower()

        # Check for exact match of full query
        if query_lower in title_lower:
            score = SCORE_TITLE
            if query_lower == title_lower:
                score = SCORE_TITLE + 5  # Bonus for exact match
            elif title_lower.startswith(query_lower):
                score = SCORE_TITLE + 2  # Bonus for prefix match
            else:
                score = SCORE_PARTIAL_TITLE

            for file_id, anchor in occurrences:
                if file_id < len(docnames):
                    docname = docnames[file_id]
                    key = (docname, anchor)
                    if key not in results or results[key][1] < score:
                        results[key] = (title_text, score)

        # Check for individual term matches in title
        else:
            matched_terms = sum(1 for term in query_terms if term.lower() in title_lower)
            if matched_terms > 0:
                # Score based on proportion of query terms found
                score = int(SCORE_PARTIAL_TITLE * (matched_terms / len(query_terms)))
                for file_id, anchor in occurrences:
                    if file_id < len(docnames):
                        docname = docnames[file_id]
                        key = (docname, anchor)
                        if key not in results or results[key][1] < score:
                            results[key] = (title_text, score)

    # 2. Search in titleterms (stemmed title keywords)
    for stemmed_term in stemmed_terms:
        if stemmed_term in titleterms:
            file_ids = titleterms[stemmed_term]
            if isinstance(file_ids, list):
                for file_id in file_ids:
                    if file_id < len(titles) and file_id < len(docnames):
                        title = titles[file_id]
                        docname = docnames[file_id]
                        key = (docname, None)
                        score = SCORE_PARTIAL_TITLE
                        if key not in results or results[key][1] < score:
                            results[key] = (title, score)

    # 3. Search in terms (full-text stemmed keywords)
    for stemmed_term in stemmed_terms:
        if stemmed_term in terms:
            file_ids = terms[stemmed_term]
            if isinstance(file_ids, list):
                for file_id in file_ids:
                    if file_id < len(titles) and file_id < len(docnames):
                        title = titles[file_id]
                        docname = docnames[file_id]
                        key = (docname, None)
                        # Only add if not already found with higher score
                        score = SCORE_TERM
                        if key not in results or results[key][1] < score:
                            results[key] = (title, score)

    return results


def format_results_sync(results, base_url, limit):
    """Format and sort search results (without summaries)."""
    formatted = []

    for (docname, anchor), (title, score) in results.items():
        url = urljoin(base_url, f"{docname}.html")
        if anchor:
            url += f"#{anchor}"

        formatted.append((title, url, score))

    formatted.sort(key=lambda x: (-x[2], x[0].lower()))

    return formatted[:limit]


async def fetch_single_summary(url, query_terms, anchor, executor):
    """Fetch and extract summary for a single result."""
    loop = asyncio.get_event_loop()

    def fetch_sync():
        try:
            with urlopen(url, timeout=10) as response:
                html_content = response.read().decode('utf-8')
                return get_summary(html_content, query_terms, anchor)
        except Exception:
            return None

    return await loop.run_in_executor(executor, fetch_sync)


async def update_result_summary(result_num, url, query_terms, anchor, result_line_counts, terminal_width, executor):
    """Fetch summary and update the display for one result."""
    summary = await fetch_single_summary(url, query_terms, anchor, executor)

    if not sys.stdout.isatty():
        if summary:
            click.echo(f"  [{result_num}] {summary}")
        return

    max_chars_per_line = terminal_width - 2
    if summary:
        # Strip ANSI codes to calculate visible length
        ansi_escape = re.compile(r'\x1b\[[0-9;]*m|\x1b\]8;;[^\x1b]*\x1b\\|\x1b\]8;;\x1b\\')
        visible_summary = ansi_escape.sub('', summary)

        has_second_line = len(visible_summary) > max_chars_per_line
        line1 = truncate_text(summary, max_chars_per_line, query_terms, add_ellipsis=not has_second_line)

        if has_second_line:
            remaining_visible = visible_summary[max_chars_per_line:]
            line2 = truncate_text(remaining_visible, max_chars_per_line, query_terms, add_ellipsis=True)
        else:
            line2 = ""
    else:
        line1 = "(no summary)"
        line2 = ""

    # Calculate lines to move back from bottom
    # We need to:
    # 1. Skip all results AFTER this one: sum(result_line_counts[result_num:])
    # 2. Skip this result's blank line: +1
    # 3. Skip this result's second summary line: +1
    # Total: sum(result_line_counts[result_num:]) + 3

    # Note: result_num is 1-indexed (1, 2, 3, ...), but result_line_counts is 0-indexed
    # result_line_counts[result_num:] correctly gives all results after this one
    lines_back = sum(result_line_counts[result_num:]) + 3

    # Move cursor to first summary line for this result
    move_cursor_up(lines_back)

    # Clear and write first summary line
    clear_line()
    sys.stdout.write(f"  {line1}")
    sys.stdout.write('\033[0m\n')

    # Clear and write second summary line
    clear_line()
    sys.stdout.write(f"  {line2}")
    sys.stdout.write('\033[0m')
    sys.stdout.flush()

    # Move cursor back to bottom
    move_cursor_down(lines_back - 1)


async def fetch_and_display_summaries(formatted_results, query_terms, base_url, result_line_counts, terminal_width):
    """Fetch summaries asynchronously and update display."""
    with ThreadPoolExecutor(max_workers=10) as executor:
        tasks = []
        for i, (title, url, score) in enumerate(formatted_results, 1):
            anchor = None
            if '#' in url:
                url_parts = url.split('#', 1)
                anchor = '#' + url_parts[1]

            task = update_result_summary(i, url, query_terms, anchor, result_line_counts, terminal_width, executor)
            tasks.append(task)

        await asyncio.gather(*tasks)


@click.command()
@click.option(
    '--url',
    help='Explicit url to documentation (searchindex.js)'
)
@click.option(
    '--repo',
    help='Repository name(s), comma-separated (e.g., "documentation,hdl,no-OS"), "all" searches all repositories.'
)
@click.option(
    '--limit',
    default=None,
    type=int,
    help='Maximum number of results to show'
)
@click.option(
    '--verbose', '-v',
    is_flag=True,
    help='Show result scores'
)
@click.option(
    '--fetch', '-f',
    default=None,
    help='Fetch file by index from previous search results, or provide a full url'
)
@click.option(
    '--format',
    type=click.Choice(['html', 'src', 'md'], case_sensitive=False),
    default='md',
    help='Output format for --fetch: html, src (source .rst/.md), md (converted markdown, default)'
)
@click.argument('query', nargs=-1, required=False)
def search(url, repo, limit, verbose, fetch, format, query):
    """Search Sphinx documentation.

    Fetches the search index from a Sphinx-generated documentation site
    and searches for the given query terms. Uses the same search algorithm
    as Sphinx's JavaScript search.

    If neither --url nor --repo is specified, searches all known repositories.

    Examples:

        # Search
        adoc search ad9084 profile
        adoc search --repo documentation -- user guide
        adoc search --repo documentation,hdl,no-OS -- axi
        adoc search --url https://analogdevicesinc.github.io/hdl/2023_R2 -- ad4630
        adoc search --repo hdl --limit 10 -- axi

        # Fetch by index from previous search
        adoc search --fetch 3                     # Default: converted markfown
        adoc search --fetch 3 --format src        # Source file (.rst or .md)
        adoc search --fetch 3 --format html       # html

        # Fetch by url
        adoc search --fetch https://analogdevicesinc.github.io/.../index.html

    Format options:
        html: Page html
        src:  Source file (.rst or .md)
        md:   Page html onverted to Markdown (default)
    """
    if fetch is not None:
        if isinstance(fetch, str) and (fetch.startswith('http://') or fetch.startswith('https://')):
            fetch_url_content(fetch, format=format)
        else:
            try:
                index = int(fetch)
                fetch_source_file(index, format=format)
            except ValueError:
                click.echo("Error: --fetch must be either an integer index or a full URL", err=True)
                click.echo(f"Got: {fetch}", err=True)
                raise click.Abort()
        return

    if not query:
        click.echo("Error: Query is required", err=True)
        raise click.Abort()

    if url and repo:
        click.echo("Error: Cannot specify both --url and --repo", err=True)
        raise click.Abort()

    if url and not url.endswith('searchindex.js'):
        if not url.endswith('/'):
            url += '/'
        url += 'searchindex.js'

    if not url and not repo:
        repo = 'all'

    if limit is None:
        limit = calculate_limit_from_terminal()

    repo_sources = []  # List of (repo_name, url) tuples
    if repo:
        if repo.strip() == 'all':
            repo_names = sorted(repos.keys())
            click.echo(f"Searching {len(repo_names)} repositories...")
        else:
            repo_names = [r.strip() for r in repo.split(',')]

            unknown_repos = [r for r in repo_names if r not in repos]
            if unknown_repos:
                click.echo(f"Error: Unknown repository(ies): {' '.join(unknown_repos)}", err=True)
                click.echo(f"Available repositories: {' '.join(sorted(repos.keys()))}", err=True)
                raise click.Abort()

        for repo_name in repo_names:
            repo_url = f"{remote_doc}{repo_name}/searchindex.js"
            repo_sources.append((repo_name, repo_url))
    else:
        repo_sources = [(None, url)]

    # At this time, don't allow multiword exact match.
    query = split_strings_in_tuple(query)
    query_str = ' '.join(query)

    stemmer = snowballstemmer.stemmer('porter')

    try:
        all_results = []

        for repo_name, repo_url in repo_sources:
            try:
                index_data = fetch_search_index(repo_url)
                if index_data is None:
                    continue

                base_url = get_base_url(repo_url)

                inventory = fetch_intersphinx_inventory(base_url)
                results = search_index(index_data, query, stemmer)

                for (docname, anchor), (title, score) in results.items():
                    result_url = urljoin(base_url, f"{docname}.html")
                    if anchor:
                        result_url += f"#{anchor}"

                    refs = get_intersphinx_references(inventory, docname, repo_name, base_url)

                    all_results.append({
                        'repo': repo_name,
                        'title': title,
                        'url': result_url,
                        'score': score,
                        'anchor': anchor,
                        'docname': docname,
                        'doc_ref': refs['doc_ref'],
                        'label_refs': refs['label_refs']
                    })

            except Exception as e:
                click.echo(f"Warning: Failed to search {repo_name or repo_url}: {e}", err=True)
                continue

        if not all_results:
            click.echo(f"\nNo results found for \"{query_str}\"")
            return

        all_results.sort(key=lambda x: (-x['score'], x['title'].lower()))
        total = len(all_results)
        click.echo(f"\nFound {total} result{'s' if total != 1 else ''} for \"{query_str}\"")

        all_results = all_results[:limit]

        formatted_results = [(r['title'], r['url'], r['score'], r['repo'], r['doc_ref'], r['label_refs'], r['docname']) for r in all_results]
        showing = len(formatted_results)

        if showing < total:
            click.echo(f"Top {showing} results:\n")
        else:
            click.echo()

        terminal_width = get_terminal_width()

        result_line_counts = []
        for i, (title, url, score, repo_name, doc_ref, label_refs, docname) in enumerate(formatted_results, 1):
            if sys.stdout.isatty():
                blue = BLUE
                reset = RESET
            else:
                blue = ""
                reset = ""

            number_prefix = f"[{i}] "

            if verbose:
                if len(repo_sources) > 1 and repo_name:
                    title_line = f"{number_prefix}{blue}[Score: {score}] [{repo_name}] {title}{reset}"
                else:
                    title_line = f"{number_prefix}{blue}[Score: {score}] {title}{reset}"
            else:
                if len(repo_sources) > 1 and repo_name:
                    title_line = f"{number_prefix}{blue}[{repo_name}] {title}{reset}"
                else:
                    title_line = f"{number_prefix}{blue}{title}{reset}"

            refs_parts = []

            if label_refs:
                clickable_refs = []
                for ref_text, label_name, full_url in label_refs:
                    clickable_ref = make_clickable_link(full_url, ref_text)
                    clickable_refs.append(clickable_ref)
                refs_parts.extend(clickable_refs)

            if doc_ref:
                clickable_doc_ref = make_clickable_link(url, doc_ref)
                refs_parts.append(clickable_doc_ref)

            if refs_parts:
                # Strip ANSI codes
                ansi_escape = re.compile(r'\x1b\[[0-9;]*m|\x1b\]8;;[^\x1b]*\x1b\\|\x1b\]8;;\x1b\\')

                max_width = terminal_width
                included_refs = []
                current_length = 0

                for ref in refs_parts:
                    visible_ref = ansi_escape.sub('', ref)
                    ref_length = len(visible_ref)

                    separator_length = 1 if included_refs else 0

                    # Check if adding this ref would exceed width
                    if current_length + separator_length + ref_length <= max_width:
                        included_refs.append(ref)
                        current_length += separator_length + ref_length
                    else:
                        # Would exceed width, stop adding refs
                        break

                # Build the final refs line
                if included_refs:
                    # If we dropped refs and " ..." doesn't fit, drop refs from the end until it does
                    if len(included_refs) < len(refs_parts):
                        ellipsis_length = 4  # len(" ...")
                        while included_refs and current_length + ellipsis_length > max_width:
                            dropped_ref = included_refs.pop()
                            visible_dropped = ansi_escape.sub('', dropped_ref)
                            if included_refs:
                                current_length -= len(visible_dropped) + 1  # +1 for " "
                            else:
                                current_length -= len(visible_dropped)

                    if included_refs:
                        refs_text = ' '.join(included_refs)
                        if len(included_refs) < len(refs_parts):
                            refs_text += " ..."
                        refs_line = click.style(refs_text, dim=True)
                    else:
                        refs_line = ""
                else:
                    refs_line = ""
            else:
                refs_line = ""

            url_line = f"{url}"

            # In non-TTY mode, summaries are printed at the bottom
            if sys.stdout.isatty():
                summary_line1 = "  ..."
                summary_line2 = "  "
                show_summary_lines = True
            else:
                summary_line1 = ""
                summary_line2 = ""
                show_summary_lines = False

            # Calculate how many physical lines this result will take
            # Structure: title + url + refs (may wrap if long) + [summary x2] + blank
            lines_count = calculate_wrapped_lines(title_line, terminal_width)
            lines_count += calculate_wrapped_lines(url_line, terminal_width)
            lines_count += calculate_wrapped_lines(refs_line, terminal_width)
            if show_summary_lines:
                lines_count += 2  # two summary lines (TTY only)
            lines_count += 1  # blank line

            result_line_counts.append(lines_count)

            click.echo(title_line)
            click.echo(url_line)
            click.echo(refs_line)
            if show_summary_lines:
                click.echo(summary_line1)
                click.echo(summary_line2)
            click.echo()

        save_search_results(formatted_results)

        formatted_for_async = [(title, url, score) for title, url, score, _, _, _, _ in formatted_results]
        asyncio.run(fetch_and_display_summaries(formatted_for_async, query, None, result_line_counts, terminal_width))

        if sys.stdout.isatty():
            click.echo()

        click.echo(click.style("Fetch source with: adoc search --fetch <index>", dim=True))

    except URLError as e:
        click.echo(f"Error: {e}", err=True)
        raise click.Abort()
    except ValueError as e:
        click.echo(f"Error: {e}", err=True)
        raise click.Abort()
    except Exception as e:
        click.echo(f"Unexpected error: {e}", err=True)
        raise click.Abort()



from .node import node_setup
from .common import common_setup
from .hdl import hdl_setup
from .system_top import system_top_setup

setup = [
    node_setup,
    common_setup,
    hdl_setup,
    system_top_setup
]



from docutils import nodes

class node_base(nodes.Element, nodes.General):
    """
    Adapted from
    https://github.com/pradyunsg/sphinx-inline-tabs
    https://github.com/dgarcia360/sphinx-collapse
    """

    @staticmethod
    def visit(self, node):
        attributes = node.attributes.copy()

        attributes.pop("ids")
        attributes.pop("classes")
        attributes.pop("names")
        attributes.pop("dupnames")
        if 'backrefs' in attributes:
            # DEPRECATED: backrefs will be removed from docutils 2.0, not in
            # python 3.14
            attributes.pop("backrefs")

        text = self.starttag(node, node.tagname, **attributes)
        self.body.append(text.strip())

    @staticmethod
    def depart(self, node):
        if node.endtag:
            self.body.append(f"</{node.tagname}>")

    @staticmethod
    def default(self, node):
        pass

class node_div(node_base):
    # prefer nodes.container
    tagname = 'div'
    endtag = 'true'

class node_input(node_base):
    tagname = 'input'
    endtag = 'false'

class node_label(node_base):
    tagname = 'label'
    endtag = 'true'

class node_icon(node_base):
    tagname = 'div'
    endtag = 'false'

class node_video(node_base):
    tagname = 'video'
    endtag = 'true'

class node_source(node_base):
    tagname = 'source'
    endtag = 'false'

class node_iframe(node_base):
    tagname = 'iframe'
    endtag = 'false'

class node_a(node_base):
    tagname = 'a'
    endtag = 'true'

class node_pre(node_base):
    tagname = 'pre'
    endtag = 'true'

class node_collection(node_base):
    """
    For now, the node is not written to the html page, but used to resolve
    metadata only.
    """
    tagname = 'div'
    endtag = 'true'

    @staticmethod
    def update_collection_env(builder, olduri, uri):
        builder.collection_image[olduri] = uri

    @staticmethod
    def depart(self, node):
        """
        Check if node has child image, if so,
        collect the uri to files.
        """
        for node_ in node:
            if isinstance(node_, nodes.image):
                oldrelfn, _ =  self.builder.env.relfn2path(
                    node['olduri'],
                    self.builder.current_docname
                )
                if node_['uri'].find('://') == -1:
                    relfn, _ =  self.builder.env.relfn2path(
                        node_['uri'],
                        self.builder.current_docname
                    )
                    if not relfn.startswith(self.builder.imagedir):
                        # likely image file not readable:
                        if node['olduri'] in self.builder.collection_image:
                            del self.builder.collection_image[node['olduri']]
                        continue
                else:
                    relfn = node_['uri']
                node_collection.update_collection_env(
                    self.builder, oldrelfn, relfn
                )
                break
        self.body.append(f"</{node.tagname}>")

def node_setup(app):
    for node in [node_div, node_input, node_label, node_icon, node_video, node_source, node_iframe, node_a, node_pre]:
        app.add_node(node,
            html =(node.visit, node.depart),
            latex=(node.default, node.default),
            text =(node.default, node.default))

    app.add_node(node_collection,
        html =(node.visit, node_collection.depart),
        latex=(node.default, node.default),
        text =(node.default, node.default))



from os import path, listdir, remove, mkdir
from os import pardir
from os import environ, stat, utime
from shutil import copy2, which, move
import click
import importlib

from sphinx.application import Sphinx

from .aux_os import aux_killpg
from .aux_git import get_git_top_level, is_git_lfs_installed, get_lfs_sha

log = {
    'no_lfs': "File .gitattributes contains lfs rules, but git-lfs is not installed.",
    'lfs_skip_smudge': "GIT_LFS_SKIP_SMUDGE=1 is set, GET requests won't smudge files, only touching the source!",
    'no_conf_py': "File conf.py not found, is {} a docs folder?",
    'inv_f': "Could not find {}, check rollup output.",
    'inv_srcdir': "Could not find SOURCEDIR {}.",
    'no_selenium': "Package 'selenium' is not installed.",
    'rollup': "Couldn't find {}, ensure this a symbolic install.",
    'node': "Couldn't find the node executable, please install nodejs.",
    'node_alt': "And the node executable is not installed.",
    'node_': """Couldn't find {}, please install the node_modules at doctools repo root, e.g.:
    npm install rollup \\
        @rollup/plugin-terser \\
        sass \\
        --save-dev""",
    'comp': "Couldn't find the minified web files.",
    'no_node_modules': "The node_modules tools are not installed to generate them.",
    'with_node_modules': "node_modules tools are installed, run with the --dev --once flags to generate them.",
    'fetch': "Do you want to fetch from the release?",
    'builder': "Unknown builder '{}', valid options are: html, pdf.",
    'no_weasyprint': "Package 'weasyprint' required for PDF generation is not installed.",
}

# Hall of shame of poorly managed artifacts
unmanaged = []
# Avoid
first_run = True
trigger_rst = ("", "")

theme_path = path.join('adi_doctools', 'theme', 'harmonic')
style_path = path.join(theme_path, 'style')
static_common_path = path.join(theme_path, 'static_common')
static_core_path = path.join(theme_path, 'static_core')
dev_pool_val = b""

BLUE = '\033[94m'
FAIL = '\033[91m'
NC = '\033[0m'

@click.command()
@click.option(
    '--directory',
    '-d',
    is_flag=False,
    type=click.Path(exists=True),
    default='.',
    help="Path to the docs folder with the Makefile."
)
@click.option(
    '--port',
    '-p',
    is_flag=False,
    type=int,
    default=8000,
    help="Port to host the docs."
)
@click.option(
    '--dev',
    '-r',
    is_flag=True,
    default=False,
    help="Watch web source code (requires symbolic install)."
)
@click.option(
    '--selenium',
    is_flag=True,
    default=False,
    help="Use selenium/Firefox instead of pooling method (html builder only)."
)
@click.option(
    '--once',
    '-o',
    is_flag=True,
    default=False,
    help="Generate the build and exit."
)
@click.option(
    '--builder',
    '-b',
    is_flag=False,
    default="html",
    help="Builder to use, valid options are: html, pdf (WeasyPrint) (default: html)."
)
def serve(directory, port, dev, selenium, once, builder):
    """
    Watch the docs and source code to rebuild it on edit.
    Two html live update strategies are available:
    Pooling: The webpage pools timestamp changes on the .dev-pool file.
    Selenium: Page reloads through Firefox's API.
    """

    import glob
    import re
    import sched
    import time
    import threading
    import signal
    import http.server
    import socketserver
    import subprocess
    import sys

    global builddir

    def symbolic_assert(file, msg):
        if not path.isfile(file):
            click.echo(msg.format(file))
            return True
        else:
            return False

    def dir_assert(file, msg):
        if not path.isdir(file):
            click.echo(msg.format(file))
            return True
        else:
            return False

    if builder not in ['html', 'pdf']:
        click.echo(log['builder'].format(builder))
        return
    if builder == 'pdf':
        environ["ADOC_MEDIA_PRINT"] = ""
        if not importlib.util.find_spec("weasyprint"):
            click.echo(log['no_weasyprint'])
            return
        from weasyprint import HTML, CSS
        from weasyprint.text.fonts import FontConfiguration
        builder = 'singlehtml'

    source_common_files = {
        'app.umd.js', 'app.umd.js.map',
        'app.min.css', 'app.min.css.map',
    }
    source_core_files = {
        'extra.umd.js', 'extra.umd.js.map',
        'extra.min.css', 'extra.min.css.map',
        'doxygen.min.css', 'doxygen.min.css.map',
        'doxygen.umd.js', 'doxygen.umd.js.map',
    }

    def signal_handler(sig, frame):
        if builder == 'html':
            click.echo("Shutting down server")
            with lock:
                shutdown_event.set()
                reload_event.set()
                http.shutdown()
                http.server_close()
            http_thread.join()
        if dev:
            aux_killpg(rollup_p)
            aux_killpg(sass_p)
        click.echo("Terminated")
        sys.exit()

    def fetch_compiled(path_):
        req = path.join(path_, 'docs', 'requirements.txt')
        dist = path.join(path_, '.dist')
        file = "adi_doctools.tar.gz"

        f = open(req, 'r')
        for line in f:
            if 'adi-doctools' in line:
                break
        f.close()

        from urllib.request import urlretrieve
        from shutil import rmtree
        if not path.isdir(dist):
            mkdir(dist)
        urlretrieve(line, path.join(dist, file))
        subprocess.call(f"tar -xf {file}",
                        shell=True, cwd=dist)
        remove(path.join(dist, file))
        for d in listdir(dist):
            break
        for f in source_common_files:
            src = path.join(dist, d, static_common_path, f)
            dest = path.join(path_, static_common_path, f)
            if not path.isdir(path.join(path_, static_common_path)):
                mkdir(path.join(path_, static_common_path))
            copy2(src, dest)
        for f in source_core_files:
            src = path.join(dist, d, static_core_path, f)
            dest = path.join(path_, static_core_path, f)
            if not path.isdir(path.join(path_, static_core_path)):
                mkdir(path.join(path_, static_core_path))
            copy2(src, dest)
        rmtree(dist)
        click.echo("Success fetching the pre-compiled files!")

    src_dir = path.abspath(path.join(path.dirname(__file__), pardir))
    par_dir = path.abspath(path.join(src_dir, pardir))
    rollup_bin = path.join(par_dir, 'node_modules', '.bin', 'rollup')
    rollup_conf = path.join(par_dir, 'ci', 'rollup.config.app.mjs')
    sass_bin = path.join(par_dir, 'node_modules', '.bin', 'sass')

    sass_conf_1 = path.join(style_path, 'app.bundle.scss') + ':' + path.join(static_common_path, 'app.min.css')
    sass_conf_2 = path.join(style_path, 'extra.bundle.scss') + ':' + path.join(static_core_path, 'extra.min.css')
    sass_conf_3 = path.join(style_path, 'doxygen.bundle.scss') + ':' + path.join(static_core_path, 'doxygen.min.css')
    sass_conf = sass_conf_1 + ' ' +  sass_conf_2 + ' ' + sass_conf_3
    if dev:
        if which("node") is None:
            click.echo(log['node'])
            sys.exit(1)
        if symbolic_assert(rollup_conf, log['rollup'].format(rollup_conf)):
            sys.exit(1)
        if symbolic_assert(rollup_bin, log['node_'].format(rollup_bin)):
            sys.exit(1)
    else:
        with_sources = True
        for s in source_common_files:
            comp_ = path.abspath(path.join(par_dir, static_common_path, s))
            if not path.isfile(comp_):
                with_sources = False
        for s in source_core_files:
            comp_ = path.abspath(path.join(par_dir, static_core_path, s))
            if not path.isfile(comp_):
                with_sources = False

        if not with_sources:
            click.echo(log['comp'])
            if which("node") is None:
                click.echo(log['node_alt'])
            elif symbolic_assert(rollup_bin, log['no_node_modules']):
                pass
            else:
                click.echo(log['with_node_modules'])
                return
            if click.confirm(log['fetch']):
                if fetch_compiled(par_dir):
                    return
            else:
                return

    if directory is None:
        click.echo("Please provide a --directory.")
        return

    with_selenium = False
    if selenium and builder == 'html':
        if importlib.util.find_spec("selenium"):
            with_selenium = True
        else:
            click.echo(log['no_selenium'])
            sys.exit(1)

    directory = path.abspath(directory)
    conf_py = path.join(directory, 'conf.py')
    if symbolic_assert(conf_py, log['no_conf_py']):
        sys.exit(1)

    builddir_ = "_build"
    builddir = path.join(directory, builddir_, builder)
    doctreedir = path.join(builddir_, "doctrees")
    sourcedir = directory
    if dir_assert(sourcedir, log['inv_srcdir']):
        sys.exit(1)

    types_lfs = []
    if git_top_level := get_git_top_level(directory):
        git_attr = path.join(git_top_level, ".gitattributes")
        if path.isfile(git_attr):
            with open(git_attr, "r") as f:
                for line in f:
                    line = line.strip()
                    if line and not line.startswith("#") and 'filter=lfs' in line:
                        match = re.match(r"\*\.([a-zA-Z0-9]+)", line)
                        if match:
                            types_lfs.append('.'+match.group(1))
            if len(types_lfs) > 0:
                if not is_git_lfs_installed():
                    click.echo(log['no_lfs'])
                    types_lfs = []
                elif ("GIT_LFS_SKIP_SMUDGE" in environ and
                      environ["GIT_LFS_SKIP_SMUDGE"] == "1"):
                    click.echo(f"{FAIL}{log['lfs_skip_smudge']}{NC}")

    # Define PDF generation
    def generate_toctree(bookmarks, indent=0):
        """
        Generate table of contents
        from: https://github.com/Kozea/WeasyPrint/issues/23#issuecomment-312447974
        """
        outline_str = ""
        for i, (label, (page, _, _), children, _) in enumerate(bookmarks, 1):
            outline_str += ('<p>%s %s <span class="page">%d</span></p>' % (
                ' ' * indent, label.lstrip('0123456789. ').rstrip('¶# '), page))
            outline_str += generate_toctree(children, indent + 2)
        return outline_str


    if builder == 'singlehtml':
        singlehtml_file = path.join(builddir, 'index.html')
        from .aux_print import sanitize_singlehtml

    def update_pdf():
        html_ = sanitize_singlehtml(singlehtml_file)

        click.echo("preparing pdf styles...")

        font_config = FontConfiguration()
        src_dir = path.abspath(path.join(path.dirname(__file__), pardir))
        css = CSS(path.join(src_dir, 'theme', 'harmonic', 'static_common', 'app.min.css'),
                  font_config=font_config)
        css_extra = CSS(path.join(src_dir, 'theme', 'harmonic', 'style', 'weasyprint.css'),
                        font_config=font_config)

        click.echo("rendering pdf content...")
        html = HTML(string=html_, base_url=path.dirname(singlehtml_file))

        document = html.render(stylesheets=[css, css_extra])

        click.echo("writing pdf...")
        document.write_pdf(path.join(builddir, '..', 'output.pdf'))
        if not once:
            click.echo("wrote pdf! waiting new user changes...")
        else:
            click.echo("wrote pdf!")

    if not with_selenium and builder == 'html':
        environ["ADOC_DEVPOOL"] = ""

    def app_subprocess_build():
        subprocess.call(f"sphinx-build -b {builder} . {builddir} -d {doctreedir} -j auto",
                        shell=True, cwd=directory)


    watch_file_src = {}
    watch_file_rst = {}
    if dev:
        w_files = []
        # Check if minified files exists, if not, run rollup once
        rollup_cache = True
        for f in source_common_files:
            if not path.isfile(f):
                rollup_cache = False
        for f in source_core_files:
            if not path.isfile(f):
                rollup_cache = False
        if not rollup_cache:
            subprocess.call(f"{rollup_bin} -c {rollup_conf}",
                            shell=True, cwd=par_dir)
            subprocess.call(f"{sass_bin} --style compressed {sass_conf}",
                            shell=True, cwd=par_dir)
        for t in ['*.umd.js*', '*.min.css*']:
            f = glob.glob(path.join(src_dir, 'theme', 'harmonic', 'static_common', t))
            w_files.extend(f)
        for t in ['*.umd.js*', '*.min.css*']:
            f = glob.glob(path.join(src_dir, 'theme', 'harmonic', 'static_core', t))
            w_files.extend(f)
        for f in w_files:
            if symbolic_assert(f, log['inv_f']):
                sys.exit(1)

        # Build doc the first time
        app_subprocess_build()
        for f in w_files:
            watch_file_src[f] = stat(f).st_mtime
        if not once:
            # Run rollup and sass in watch mode
            cmd_rollup = f"{rollup_bin} -c {rollup_conf} --watch"
            cmd_sass = f"{sass_bin} --style compressed --watch {sass_conf}"
            rollup_p = subprocess.Popen(cmd_rollup, shell=True, cwd=par_dir,
                                        stdout=subprocess.DEVNULL)
            sass_p = subprocess.Popen(cmd_sass, shell=True, cwd=par_dir,
                                      stdout=subprocess.DEVNULL)
        elif builder == "singlehtml":
            update_pdf()
    else:
        # Build doc the first time
        app_subprocess_build()
        if builder == "singlehtml":
            update_pdf()

    if once:
        return

    # app.build() doesn't handle the cache well in parallel,
    # instead, we call through subprocess if needed
    app = Sphinx(directory, directory,  builddir,
                 doctreedir, builder, parallel=0)

    def get_source_lfs_file(path_, ext):
        """
        Check if _build binary file is a git lfs pointer,
        and if so, search and return the source file path.
        If any step fails, returns None
        """
        if sha := get_lfs_sha(path_):
            name_ = path.basename(path_)
            name_ = re.sub(r'(_?\d+)?\.\w+$', '', name_)
            pattern = path.join(directory, '**', f'*{ext}')
            files = []
            for file_path in glob.glob(pattern, recursive=True):
                if ('/_build/' not in path.normpath(file_path) and
                    name_ in path.basename(file_path)):
                    files.append(file_path)

            for file in files:
                try:
                    with open(file, 'rb') as f:
                        f.seek(54)
                        sha_ = f.read(64)
                        if sha == sha_:
                            return file
                except Exception:
                    pass

            return None
        else:
            return None

    builddir = path.join(directory, builddir_, builder)
    shutdown_event = threading.Event()
    reload_event = threading.Event()
    dev_pool_lock = threading.Lock()
    class Handler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            global builddir
            super().__init__(*args, directory=builddir, **kwargs)

        def do_GET(self):
            global dev_pool_val

            if shutdown_event.is_set():
                return

            def send_dev_pool():
                self.send_response(200)
                self.send_header("Content-type", "text/plain")
                self.end_headers()
                with dev_pool_lock:
                    self.wfile.write(dev_pool_val)
                self.end_headers()
                return

            try:
                if not with_selenium and self.path == "/.dev-pool":
                    if self.headers.get("no-wait") is not None:
                        send_dev_pool()
                        return
                    if reload_event.wait(timeout=30):
                        reload_event.clear()
                        send_dev_pool()
                    else:
                        with dev_pool_lock:
                            dev_pool_val_ = f"{str(time.time())}\n@timed-out\n"
                            dev_pool_val = bytes(dev_pool_val_, 'utf-8')
                        send_dev_pool()
                    return

                for ext in types_lfs:
                    if self.path.endswith(ext):
                        path_ = path.join(builddir, self.path[1:])

                        lfs_f = get_source_lfs_file(path_, ext)
                        if lfs_f is not None:
                            tmp_f = path.join(directory, builddir_, path.basename(lfs_f))
                            stat_ = stat(lfs_f)
                            lfs_f_ = path.relpath(lfs_f, git_top_level)
                            click.echo(f"git lfs smudging file {lfs_f_}")
                            try:
                                with open(path_, "rb") as fin, open(tmp_f, "wb") as fout:
                                    subprocess.run(["git", "lfs", "smudge"], stdin=fin, stdout=fout, check=True)
                            except Exception as e:
                                if e.returncode == 2:
                                    pass
                                else:
                                    raise
                            copy2(tmp_f, lfs_f)
                            utime(lfs_f, (stat_.st_atime, stat_.st_mtime))
                            move(tmp_f, path_)
                            utime(path_, (stat_.st_atime, stat_.st_mtime))

                            while True:
                                try:
                                    subprocess.run(["git", "update-index", "--", lfs_f],
                                                   capture_output=True, text=True, check=True)
                                except subprocess.CalledProcessError as e:
                                    if e.returncode == 128:
                                        if ".git/index.lock" in e.stderr:
                                            time.sleep(0.1)
                                            continue
                                        else:
                                            raise
                                    else:
                                        raise

                                break

                super().do_GET()
            except (BrokenPipeError, ConnectionResetError):
                return

        def log_message(self, format, *args):
            return

    def wsl2_networking(scheduler):
        """
        Helps Windows pick-up WSL2 exposed port.
        """
        try:
            from urllib.request import urlopen
            urlopen(f"http://0.0.0.0:{port}")
        except Exception as e:
            click.echo(e)

    if builder == "html":
        try:
            http = socketserver.ThreadingTCPServer(("", port), Handler)
            lock = threading.Lock()
            http_thread = threading.Thread(target=http.serve_forever)
            http_thread.daemon = True
            http_thread.start()
            click.echo(f"\n{BLUE}Running server on http://0.0.0.0:{port}{NC}\n")

            scheduler = sched.scheduler(time.time, time.sleep)
            scheduler.enter(1, 1, wsl2_networking, (scheduler,))
            scheduler.run()
        except Exception:
            click.echo(f"{FAIL}Could not start server on http://0.0.0.0:{port}{NC}")
            click.echo(f"  {BLUE}Tip{NC}: pass another port with {BLUE}--port{NC}")
            if dev:
                aux_killpg(rollup_p)
                aux_killpg(sass_p)
            return
    signal.signal(signal.SIGINT, signal_handler)

    dev_pool = path.join(builddir, '.dev-pool')

    def update_dev_pool(file):
        global dev_pool_val
        dev_pool_val_ = f"{str(time.time())}\n{file}\n"
        # For XHR
        with dev_pool_lock:
            dev_pool_val = bytes(dev_pool_val_, 'utf-8')
        reload_event.set()
        # For alt servers, e.g. at ../ of {./doctools, ./no-OS}
        if path.isdir(builddir):
            dev_f = open(dev_pool, 'w')
            dev_f.write(dev_pool_val_)
            dev_f.close()

    if with_selenium:
        from selenium import webdriver

        # Remove pooling method flag
        if path.isfile(dev_pool):
            remove(dev_pool)

        driver = webdriver.Firefox()

        driver.get(f"http://0.0.0.0:{port}")
    elif builder == "html":
        update_dev_pool("")


    def get_doc_sources_included():
        include_ = set()
        for docname in app.env.included:
            include_.update(app.env.included[docname])
        files_ = {item + ".rst" for item in include_}
        files = []
        ctime = []
        for f in files_:
            if not path.isfile(f):
                continue
            ctime_ = stat(f).st_mtime
            ctime.append(ctime_)
            files.append(f)
        return (files, ctime)


    def get_doc_sources():
        types = ['*.rst', '*.md', '*.svg', '*.txt', '*.png', '*.jpg', '*.jpeg', '*.js', '*.css']
        files = []
        ctime = []
        for typ in types:
            glob_ = path.join(sourcedir, typ)
            _files = glob.glob(glob_)
            __files = [path.abspath(f) for f in _files]
            files.extend(__files)
            for f in __files:
                ctime.append(stat(f).st_mtime)
        if path.isfile(f):
            files.append(conf_py)
            ctime.append(stat(conf_py).st_mtime)
        dirs = [d for d in listdir(sourcedir)
                if path.isdir(path.join(sourcedir, d))]
        if builddir_ in dirs:
            dirs.remove(builddir_)
        for d in dirs:
            for typ in types:
                glob_ = path.join(sourcedir, d, '**', typ)
                _files = glob.glob(glob_, recursive=True)
                __files = [path.abspath(f) for f in _files]
                for f in __files:
                    if not path.isfile(f):
                        continue
                    ctime_ = stat(f).st_mtime
                    ctime.append(ctime_)
                    files.append(f)
        files_, ctime_ = get_doc_sources_included()
        files.extend(files_)
        ctime.extend(ctime_)
        return (files, ctime)


    def get_trigger_rst(trigger_rst_, file):
        if trigger_rst_[0] == file:
            return trigger_rst_

        if file.endswith(".rst"):
            c = -4
        elif file.endswith(".md"):
            c = -3
        else:
            return trigger_rst_

        path_ = path.relpath(file, sourcedir)[:c] + ".html"
        if path_.startswith("../"):
            # Outside of source dir, unsupported
            return trigger_rst_
        return (file, path_)


    def check_files(scheduler):
        global first_run, trigger_rst
        update_sphinx = False
        update_dev = False
        git_lfs_pull = []
        for file, ctime in zip(*get_doc_sources()):
            if file in watch_file_rst and ctime > watch_file_rst[file]:
                _, ext_= path.splitext(file)
                if ext_ in types_lfs and get_lfs_sha(file):
                    # User touched lfs symbolic link, probably wants to pull it
                    git_lfs_pull.append(file)

            if file in watch_file_rst and ctime > watch_file_rst[file]:
                trigger_rst = get_trigger_rst(trigger_rst, file)
            if file not in watch_file_rst or ctime > watch_file_rst[file]:
                update_sphinx = True
                watch_file_rst[file] = ctime
                for u in unmanaged:
                    if u in file:
                        watch_file_rst[file] = sys.float_info.max
                        update_sphinx = False
                        break

        for file in watch_file_src:
            if not path.isfile(file):
                continue
            ctime = stat(file).st_mtime
            if ctime > watch_file_src[file]:
                update_dev = True
                watch_file_src[file] = ctime

        use_subprocess = False
        if not path.isdir(builddir):
            # User did make clean
            update_sphinx = True
            use_subprocess = True

        if first_run is True:
            first_run = False
            update_dev = False
            update_sphinx = False

        if len(git_lfs_pull) > 0:
            git_lfs_pull = [path.relpath(gf, git_top_level) for gf in git_lfs_pull]
            lfs_f_s = ' -I '.join(git_lfs_pull)
            click.echo(f"git lfs smudging file(s): {' '.join(git_lfs_pull)}")
            try:
                subprocess.run(["git", "lfs", "pull", "-I", lfs_f_s], check=True,
                                cwd=git_top_level)
            except Exception as e:
                if e.returncode == 2:
                    pass
                else:
                    raise

        if update_sphinx:
            # dev:
            #   Uses subprocess because creating a new Sphinx class:
            #   * Do not re-eval the roles/directives, but
            #   * Trigger a full rebuild due to env changes
            #   so it is no use for developing purposes.
            #
            #   Maybe importlib.reload() + monkey patch could be an alternative,
            #   but not triggering full env reload would be tricky, so this is
            #   good enough.
            if dev or use_subprocess:
                app_subprocess_build()
            else:
                app.build()
        if update_dev:
            for f in w_files:
                copy2(f, path.join(builddir, '_static', path.basename(f)))
        if update_sphinx or update_dev:
            if with_selenium:
                try:
                    driver.execute_script(f"location.replace('{trigger_rst[1]}');")
                except Exception:
                    click.echo("Browser disconnected")
                    if dev:
                        aux_killpg(rollup_p)
                        aux_killpg(sass_p)
                    with lock:
                        shutdown_event.set()
                        reload_event.set()
                        http.shutdown()
                        http.server_close()
                    http_thread.join()
                    return
            elif builder == "html":
                update_dev_pool("@code-changed" if update_dev else trigger_rst[1])
            elif builder == 'singlehtml':
                update_pdf()

        scheduler.enter(1, 1, check_files, (scheduler,))

    scheduler = sched.scheduler(time.time, time.sleep)
    scheduler.enter(1, 1, check_files, (scheduler,))
    scheduler.run()




class string_system_top():
    esd_warning = '''All the products described on this page include ESD (electrostatic discharge) sensitive devices. Electrostatic charges as high as 4000V readily accumulate on the human body or test equipment and can discharge without detection.
Although the boards feature ESD protection circuitry, permanent damage may occur on devices subjected to high-energy electrostatic discharges. Therefore, proper ESD precautions are recommended to avoid performance degradation or loss of functionality. This includes removing static charge on external equipment, cables, or antennas before connecting to the device. '''

class string_hdl ():
    access_type = {
        'RO':{
            'name': 'Read-only',
            'description': 'Reads will return the current register value. Writes have no effect.'
        },
        'WO':{
            'name': 'Write-only',
            'description': 'Writes will change the current register value. Reads have no effect.'
        },
        'RW':{
            'name': 'Read-write',
            'description': 'Reads will return the current register value. Writes will change the current register value.'
        },
        'RW1C':{
            'name': 'Read,write-1-to-clear',
            'description': 'Reads will return the current register value. Writing the register will clear those bits of the register which were set to 1 in the value written. Bits are set by hardware.'
        },
        'RW1S':{
            'name': 'Read,write-1-to-set',
            'description': 'Reads will return the current register value. Writing the register will set those bits of the register which were set to 1 in the value written. Bits are cleared by hardware.'
        },
        'W1S':{
            'name': 'Write-1-to-set',
            'description': 'Writing the register will set those bits of the register which were set to 1 in the value written. Bits are cleared by hardware. Reads have no effect.'
        },
        'V':{
            'name': 'Volatile',
            'description': 'The V suffix indicates that the register is volatile and its content might change without software interaction. The value of registers without the volatile designation will not change without an explicit write done by software.'
        },
        'NA':{
            'name': 'No access',
            'description': 'Do not read or write the register.'
        }
    }



"""
Extends intersphinx.
"""
from typing import Dict, Any

from os import path, getenv
from packaging.version import Version
from sphinx.__init__ import __version__ as __sphinx_version__
from sphinx.util.osutil import SEP

from ..lut import repos, remote_doc

# For deprecation handling
from sphinx.util import logging
from sphinx.application import Sphinx

logger = logging.getLogger(__name__)

dft_interref_uri = remote_doc


def interref_repos_apply(config):
    """
    Applies interref_repos into intersphinx_mapping.
    Expands the repository name into targets, with the main one being:
    interref_uri + repo, e.g.
        docs.app.com/my-repo
    If interref_local is True, add
    a secondary relative to the current doc considering the target repo
    is alongside the current doc, e.g.
    - /data/my-repo-0/docs
      /data/my-repo-1/doc/sphinx/source
    Useful for when simultaneously editing multiple docs.
    """
    if config.interref_uri is None:
        config.interref_uri = getenv("ADOC_INTERREF_URI",
                                     default=dft_interref_uri)

    def repo_apply_(r):
        if r in repos:
            d_ = (f'..{SEP}'*(repos[r]['pathname'].count(SEP)+1) or f'..{SEP}')
            if 'parent' in repos[r] and repos[r]['parent'] is not None:
                d_ += f'..{SEP}'
            return d_ + f'..{SEP}'

    t_ = None
    if 'repository' in config and config.interref_local:
        t_ = repo_apply_(config.repository)
    for r in config.interref_repos:
        r_ = r.split('/')[0]
        if t_ is not None and r_ in repos:
            t__ = path.join(t_, r, repos[r_]['pathname'], '_build',
                            'html', 'objects.inv')
            t__ = path.abspath(t__)
            t__ = (t__, None)
        else:
            t__ = None
        config.intersphinx_mapping[r_] = (config.interref_uri+r, t__)
    config.interref_repos = []  # Consumed


def interref_repos_assert(config):
    if len(config.interref_repos):
        logger.warning("interref_repos is present in config but never applied."
                       " Is adi_doctools in the extension list?")


def interref_setup(app: Sphinx) -> Dict[str, Any]:
    app.add_config_value('interref_repos', [], 'env')
    app.add_config_value('interref_uri', None, 'env')
    app.add_config_value('interref_local', False, 'env')

    if not isinstance(app.config.interref_repos, list):
        logger.warning("Config 'interref_repos' must be a list.")
        app.config.interref_repos = []

    if Version(__sphinx_version__) < Version("7.3.0"):
        # Prior to Sphinx 7.3.x, config values were default during setup,
        # so we cannot enable on demand.
        auto_enable_intersphinx = True
    elif len(app.config.interref_repos) > 0:
        auto_enable_intersphinx = True
    else:
        auto_enable_intersphinx = False

    if auto_enable_intersphinx:
        app.setup_extension('sphinx.ext.intersphinx')
        if (len(app.config.intersphinx_disabled_reftypes) == 1 and
           app.config.intersphinx_disabled_reftypes[0] == 'std:doc'):
            app.config.intersphinx_disabled_reftypes = ["*"]
            logger.info("adi_doctools: "
                        "Changed intersphinx_disabled_reftypes value from "
                        "'std:doc' to '*' to avoid refs linking to wrong docs")



from typing import Dict

from datetime import datetime
from os import path

from ..__init__ import __version__
from ..typing.hdl import Library, Project

license_makefile = f"""\
####################################################################################
## Copyright (c) 2018-{datetime.now().year} Analog Devices, Inc.
### SPDX short identifier: BSD-1-Clause
## Auto-generated v{__version__}, do not modify!
####################################################################################
"""

license_sv = """\
// ***************************************************************************
// ***************************************************************************
// Copyright (C) 2014-2025 Analog Devices, Inc. All rights reserved.
//
// In this HDL repository, there are many different and unique modules, consisting
// of various HDL (Verilog or VHDL) components. The individual modules are
// developed independently, and may be accompanied by separate and unique license
// terms.
//
// The user should read each of these license terms, and understand the
// freedoms and responsibilities that he or she has by using this source/core.
//
// This core is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.
//
// Redistribution and use of source or resulting binaries, with or without modification
// of this file, are permitted under one of the following two license terms:
//
//   1. The GNU General Public License version 2 as published by the
//      Free Software Foundation, which can be found in the top level directory
//      of this repository (LICENSE_GPL2), and also online at:
//      <https://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
//
// OR
//
//   2. An ADI specific BSD license, which can be found in the top level directory
//      of this repository (LICENSE_ADIBSD), and also on-line at:
//      https://github.com/analogdevicesinc/hdl/blob/main/LICENSE_ADIBSD
//      This will allow to generate bit files and not release the source code,
//      as long as it attaches to an ADI device.
//
// ***************************************************************************
// ***************************************************************************
"""

def svpkg_regmap(f, regmap: Dict, key: str):
    f.write(f"    /* {regmap['title']} */\n")

    for reg in regmap['regmap']:
        # Skip unresolved
        if reg['import']:
            continue
        row = f"    class {reg['name']}_CLASS"
        row += " extends register_base;""\n"
        f.write(row)

        for field in reg['fields']:
            if field['name'] != 'RESERVED' and not field['import']:
                row = f"      field_base {field['name']}_F;""\n"
                f.write(row)

        f.write("\n      function new(\n")
        f.write("        input string name,\n")
        f.write("        input int address,\n")
        reg_param_dec = []
        for reg_param in reg['parameters']:
            reg_param_dec.append(reg_param)
        if len(reg_param_dec):
            reg_params_set = set(reg_param_dec)
            reg_param_dec = list(reg_params_set)
            reg_param_dec.sort()
            for reg_param in reg_param_dec:
                f.write(f"        input int {reg_param},\n")
        f.write("        input adi_regmap parent = null);\n\n")
        f.write("        super.new(name, address, parent);\n\n")
        for field in reg['fields']:
            if field['name'] != 'RESERVED' and not field['import']:
                row = f"        this.{field['name']}_F = "'new("'f"{field['name']}"
                bits = f"{field['bits'][0]}, {field['bits'][1]}"

                if field['default_long'] is None:
                    default = "'hXXXXXXXX"
                else:
                    default = field['default_long']
                    if type(default) is int:
                        default = hex(field['default_long']).replace("0x", "'h")

                row += '"'f", {bits}, {field['rw']}, {default}, this);""\n"
                f.write(row)

        row = "\n        this.initialization_done = 1;""\n"
        f.write(row)

        f.write("      endfunction: new\n")
        row = f"    endclass: {reg['name']}_CLASS\n\n"
        f.write(row)


def svpkg_head(f, key: str, regmap: Dict):
    run_time = datetime.now().strftime('%b %d %H:%M:%S %Y')
    pkgname = f"adi_regmap_{key}_pkg"
    classname = f"adi_regmap_{key}"
    f.write(license_sv)
    f.write("/* Auto generated Register Map */\n")
    f.write(f"/* {run_time} v{__version__} */\n")
    f.write("\n")

    f.write(f"package {pkgname};\n")
    f.write("  import logger_pkg::*;\n")
    f.write("  import adi_api_pkg::*;\n\n")
    f.write(f"  class {classname} extends adi_regmap")
    f.write(";\n\n")


def svpkg_reg_decl(f, regmap: Dict):
    for reg in regmap['regmap']:
        if reg['where'] is not None:
            number_of_instances = reg['where'][1]-1
            row = f"    {reg['name']}_CLASS"
            row += f" {reg['name']}_R [{number_of_instances}:0];\n"
            f.write(row)
        else:
            row = f"    {reg['name']}_CLASS"
            row += f" {reg['name']}_R;\n"
            f.write(row)


def svpkg_reg_inst(f, regmap: Dict):
    for reg in regmap['regmap']:
        if reg['where'] is not None:
            number_of_instances = reg['where'][1]
            f.write(f"      for (int i=0; i<{number_of_instances}; i++) begin\n")
            split_name = reg['name'].split('n')
            reg['address'] = (reg['address'] + reg['addr_incr'] * 0) * 4
            addr = hex(reg['address']).replace("0x", "'h")
            row = f"        this.{reg['name']}_R[i] = new("
            row += f"$sformatf(\"{split_name[0]}%0d{split_name[1]}\", i)"
            row += f", {addr} + 'h{reg['addr_incr']} * i * 4"
            reg_param_dec = []

            for reg_param in reg['parameters']:
                reg_param_dec.append(reg_param)
            if len(reg_param_dec):
                reg_params_set = set(reg_param_dec)
                reg_param_dec = list(reg_params_set)
                reg_param_dec.sort()
                row += ", "
                row += ", ".join(reg_param_dec)

            row += ", this);\n"
            f.write(row)
            f.write("      end\n")
        else:
            reg['address'] = reg['address'] * 4
            addr = hex(reg['address']).replace("0x", "'h")
            row = f"      this.{reg['name']}_R = new("
            row += '"' + reg['name'] + '"'
            row += f", {addr}"
            reg_param_dec = []
            for reg_param in reg['parameters']:
                reg_param_dec.append(reg_param)
            if len(reg_param_dec):
                reg_params_set = set(reg_param_dec)
                reg_param_dec = list(reg_params_set)
                reg_param_dec.sort()
                row += ", "
                row += ", ".join(reg_param_dec)
            row += ", this);\n"
            f.write(row)


def svpkg_footer(f, key: str, regmap: Dict):
    pkgname = f"adi_regmap_{key}_pkg"
    classname = f"adi_regmap_{key}"
    f.write(f"  endclass: {classname}\n\n")
    f.write(f"endpackage: {pkgname}\n")


def write_hdl_regmap_pkg(
    path_: str,
    regmap: Dict,
    key: str
) -> None:
    fname = f"adi_regmap_{key}_pkg.sv"
    file = path.join(path_, fname)
    f = open(file, "w")
    svpkg_head(f, key, regmap)

    for rm in regmap:
        svpkg_regmap(f, regmap[rm], rm)

    for rm in regmap:
        svpkg_reg_decl(f, regmap[rm])

    f.write("\n    function new(\n")
    f.write("      input string name,\n")
    f.write("      input int address,\n")
    reg_param_dec = []
    for rm in regmap:
        for reg in regmap[rm]['regmap']:
            for reg_param in reg['parameters']:
                reg_param_dec.append(reg_param)
    if len(reg_param_dec):
        reg_params_set = set(reg_param_dec)
        reg_param_dec = list(reg_params_set)
        reg_param_dec.sort()
        for reg_param in reg_param_dec:
            f.write(f"      input int {reg_param},\n")
    f.write("      input adi_api parent = null);\n\n")
    f.write("      super.new(name, address, parent);\n\n")
    for rm in regmap:
        svpkg_reg_inst(f, regmap[rm])
    f.write("\n")
    f.write("      this.info($sformatf(\"Initialized\"), ADI_VERBOSITY_HIGH);\n")
    f.write("    endfunction: new\n\n")

    svpkg_footer(f, key, rm)

    f.close()


def write_hdl_regmap_definitions(
    path_: str,
    regmap: Dict,
    key: str
) -> None:
    reg_param_dec = []
    for rm in regmap:
        for reg in regmap[rm]['regmap']:
            for reg_param in reg['parameters']:
                reg_param_dec.append(reg_param)
    if len(reg_param_dec):
        reg_params_set = set(reg_param_dec)
        reg_param_dec = list(reg_params_set)
        reg_param_dec.sort()
        separator = ', \\\n'

        fname = f"adi_regmap_{key}_definitions.svh"
        file = path.join(path_, fname)
        f = open(file, "w")

        run_time = datetime.now().strftime('%b %d %H:%M:%S %Y')
        pkgname = f"adi_regmap_{key}_pkg"
        f.write(license_sv)
        f.write("/* Auto generated Register Map */\n")
        f.write(f"/* {run_time} v{__version__} */\n")
        f.write("\n")

        f.write("`timescale 1ns/1ps\n\n")
        f.write(f"`ifndef _{pkgname.upper()}_DEFINITIONS_SVH_\n")
        f.write(f"`define _{pkgname.upper()}_DEFINITIONS_SVH_\n\n")

        f.write("// Help build VIP Interface parameters name\n")

        reg_param_dec_import = reg_param_dec.copy()
        for i, reg_param in enumerate(reg_param_dec_import):
            reg_param_dec_import[i] = f"  n``.inst.{reg_param}"
        row = separator.join(reg_param_dec_import)
        f.write(f"`define {pkgname.upper()}_PARAM_IMPORT(n)")
        f.write(f"{row}\n\n")

        for i, reg_param in enumerate(reg_param_dec):
            reg_param_dec[i] = f"  {reg_param}"
        row = separator.join(reg_param_dec)
        f.write(f"`define {pkgname.upper()}_PARAM_DECL int")
        f.write(f"{row}\n\n")

        f.write(f"`define {pkgname.upper()}_PARAM_ORDER")
        f.write(f"{row}\n\n")

        f.write("`endif\n")

        f.close()


def write_hdl_regmap(
    path_: str,
    regmap: Dict,
    key: str
) -> None:
    write_hdl_regmap_pkg(path_, regmap, key)
    write_hdl_regmap_definitions(path_, regmap, key)


def regmap_test_program(
    path_: str,
    regmap: Dict
) -> None:
    fname = "test_program.sv"
    file = path.join(path_, fname)
    f = open(file, "w")

    f.write("`include \"utils.svh\"\n\n")
    f.write("import logger_pkg::*;\n")

    for m in regmap:
        row = "import adi_regmap_" + m + "_pkg::*;\n"
        f.write(row)

    f.write("\nmodule test_program;\n\n")

    for m in regmap:
        row = "  adi_regmap_" + m + " "

        row += "adi_regmap_" + m + "_rm;\n"
        f.write(row)

    f.write("\n  initial begin\n\n")

    f.write("\n    setLoggerVerbosity(ADI_VERBOSITY_NONE);\n\n")

    for m in regmap:
        row = "    adi_regmap_" + m + "_rm = new(\"" + m + "\""
        row += ", 0"
        reg_param_dec = []
        for k in regmap[m]['subregmap']:
            for reg in regmap[m]['subregmap'][k]['regmap']:
                for reg_param in reg['parameters']:
                    reg_param_dec.append(reg_param)
        if len(reg_param_dec):
            reg_params_set = set(reg_param_dec)
            reg_param_dec = list(reg_params_set)
            reg_param_dec.sort()
            reg_param_t0 = []
            for conv in reg_param_dec:
                reg_param_t0.append("0")
            row += ", "
            row += ", ".join(reg_param_t0)
        row += ");\n"
        f.write(row)

    f.write("\n    $finish();\n\n")
    f.write("  end\n\n")
    f.write("endmodule\n")

    f.close()


def regmap_bash_script(
    path_: str,
    regmap: Dict
) -> None:
    fname = "test_program"
    file = path.join(path_, fname)
    f = open(file, "w")

    f.write("#!/bin/bash\n\n")

    f.write("SOURCE=\"utils.svh \"\n")
    f.write("SOURCE+=\"logger_pkg.sv \"\n")
    f.write("SOURCE+=\"adi_common_pkg.sv \"\n")
    f.write("SOURCE+=\"adi_environment_pkg.sv \"\n")
    f.write("SOURCE+=\"adi_vip_pkg.sv \"\n")
    f.write("SOURCE+=\"axi_definitions.svh \"\n")
    f.write("SOURCE+=\"m_axi_sequencer.sv \"\n")
    f.write("SOURCE+=\"adi_api_pkg.sv \"\n")

    for m in regmap:
        row = "SOURCE+=\"adi_regmap_" + m + "_pkg.sv \"\n"
        f.write(row)

    f.write("SOURCE+=\"test_program.sv\"\n\n")

    f.write("cp ../utilities/utils.svh .\n")
    f.write("cp ../utilities/logger_pkg.sv .\n")
    f.write("cp ../utilities/adi_common_pkg.sv .\n")
    f.write("cp ../utilities/adi_environment_pkg.sv .\n")
    f.write("cp ../utilities/adi_api_pkg.sv .\n")
    f.write("cp ../utilities/adi_vip_pkg.sv .\n")
    f.write("cp ../vip/amd/axi/m_axi_sequencer.sv .\n")
    f.write("cp ../vip/amd/axi/axi_definitions.svh .\n")
    f.write("source ../../scripts/run_unit_tb.sh")

    f.close()


def write_hdl_regmap_test(
    path_: str,
    regmap: Dict
) -> None:
    regmap_test_program(path_, regmap)
    regmap_bash_script(path_, regmap)


def write_hdl_library_makefile(
    libraries: Dict[str, Library],
    path_: str
) -> None:
    library = libraries[path_]
    fname = "Makefile"
    file = path.join('library', path_, fname)
    f = open(file, "w")
    f.write(license_makefile)
    f.write("\n")
    f.write(f"LIBRARY_NAME := {library['name']}\n")
    f.write("\n")
    for d in library['generic']['dependencies']:
        f.write(f"GENERIC_DEPS += {d}\n")
    f.write("\n")
    for v in library['vendor']:
        r = library['vendor'][v]
        for d in r['dependencies']:
            f.write(f"{v.upper()}_DEPS += {d}\n")
        if len(r['dependencies']) > 0:
            f.write("\n")

        for x in r['interfaces']:
            f.write(f"{v.upper()}_DEPS += {x}\n")
        if len(r['interfaces']) > 0:
            f.write("\n")

        for d in r['library_dependencies']:
            f.write(f"{v.upper()}_LIB_DEPS += {d}\n")
        if len(r['library_dependencies']) > 0:
            f.write("\n")

        for x in r['interfaces_tcl']:
            f.write(f"{v.upper()}_INTERFACE_DEPS += {x}\n")
        if len(r['interfaces_tcl']) > 0:
            f.write("\n")

    p_ = path.join('scripts', 'library.mk')
    p_ = path.relpath(p_, path_)
    f.write(f"include {p_}\n")
    f.close()


def write_hdl_project_makefile(
    project: Dict[str, Project],
    path_: str
) -> None:
    project = project[path_]
    fname = "Makefile"
    file = path.join('projects', path_, fname)
    f = open(file, "w")
    f.write(license_makefile)
    f.write("\n")
    f.write(f"PROJECT_NAME := {project['name']}\n")
    f.write("\n")
    for d in project['m_deps']:
        f.write(f"M_DEPS += {d}\n")
    f.write("\n")
    for d in project['lib_deps']:
        f.write(f"LIB_DEPS += {d}\n")
    f.write("\n")
    p_ = f"scripts/project-{project['vendor']}.mk"
    p_ = path.relpath(p_, path_)
    f.write(f"include {p_}\n")
    f.close()



import os
import click

from ..parser.hdl import parse_hdl_component
from ..writer.hdl_component import hdl_component


@click.command()
@click.option(
    '--input',
    '-i',
    'input_',
    is_flag=False,
    type=click.Path(exists=True),
    default=None,
    required=True,
    help="Path to the library folder."
)
@click.option(
    '--output',
    '-o',
    is_flag=False,
    type=click.Path(exists=False),
    default="",
    help="Output path, defaults to --input_."
)
@click.option(
    '--open',
    '-x',
    'open_',
    is_flag=True,
    default=False,
    help="Open after generation (xdg-open)."
)
def hdl_render(input_, output, open_):
    """
    Creates a SVG component diagram of an IP.
    Requires the component.xml to be generated first.
    """
    input_ = os.path.abspath(input_)
    if output == "":
        output = input_

    file = os.path.join(input_, "component.xml")
    out_file = os.path.join(output, "component.svg")

    lib_name = os.path.basename(input_)
    if not os.path.isfile(file):
        click.echo(f"Component {file} not found!")
        if open_:
            tree = hdl_component.render_placeholder(file)
        else:
            return
    else:
        lib = parse_hdl_component(file, os.path.getctime(file))
        tree = hdl_component.render(lib_name, lib)

    if not os.path.exists(output):
        os.makedirs(output)
    tree.write(out_file)

    if open_:
        import subprocess

        subprocess.call(f"xdg-open {out_file}", shell=True)



from typing import TypedDict, Optional, Tuple, Dict

vendors = ("xilinx", "intel")


class IntfPort(TypedDict):
    direction: str
    width: int
    name: str
    domain: Optional[str]
    default: Optional[int]


class Intf(TypedDict):
    description: Optional[str]
    name: str
    ports: Tuple[IntfPort]


class LibraryGeneric(TypedDict):
    dependencies: Tuple[str]


class LibraryVendor(TypedDict):
    dependencies: Tuple[str]
    library_dependencies: Tuple[str]
    interfaces: Tuple[str]
    interfaces_tcl: Tuple[str]
    parameters: Optional[Tuple[Tuple[str, str]]]


class Library(TypedDict):
    """
    __key__: xilinx/some_ip
    name: adi_jesd204_glue, axi_spi_engine
    """
    name: str
    vendor: Dict[str, LibraryVendor]
    generic: LibraryGeneric


class Carrier(TypedDict):
    xilinx: Tuple[str]
    intel: Tuple[str]


class Project(TypedDict):
    """
    __key__: some_project/carrier, some_project
    name: some_project_carrier, some_project
    lib_deps: my_ip, framework/framework_core
    m_deps: some_constr.xdc
    """
    name: str
    vendor: str
    lib_deps: Tuple[str]
    m_deps: Tuple[str]



from typing import List, Set, Optional, Union, Tuple

import re
from os import path

from sphinx.util import logging

logger = logging.getLogger(__name__)

class tcl:
    def __init__(self, file: str):
        """
        Tcl systax is line break sensitive -> squash escaped line breaks on
        open to simply parsing later.
        Tabs are the same as spaces -> replace all tabs with space.
        Strip leading whitespace and from common methods like lists.
        """
        data = []
        with open(file, "r") as f:
            line_ = ''
            for line in f:
                if line.endswith('\\\n'):
                    line_ += line[:-2]
                elif any(line.lstrip().startswith(x) for x in [']', '}']) and line_ == '':
                    data[-1] = data[-1] + ' ' + line[:-1]
                else:
                    line_ += line.replace('\n', '')
                    data.append(line_)
                    line_ = ''
        data = [' '.join(d.replace("\t", " ").strip().split()) for d in data]
        for m in ["list", "info"]:
            data = [d.replace("[ "+m, "["+m) for d in data]
        self.data = data

    def __iter__(self):
        return iter(self.data)

    @staticmethod
    def get_list_items(
        line: str
    ) -> List[str]:
        i1 = line.find("[list")
        if i1 != -1:
            i2 = line.index("]")
            i_ = 5
        else:
            i1 = line.index("{")
            i2 = line.index("}")
            i_ = 1

        if i2 < i1:
            return []

        items = line[i1+i_:i2].split()
        for i, item in enumerate(items):
            if item[0] == '"' and item[-1] == '"':
                items[i] = item[1:-1]

        return items

    def line_startswith(
        self,
        start: Union[List[str], str]
    ) -> Optional[str]:
        start = [start] if type(start) is str else start
        for line in self.data:
            for s in start:
                if line.startswith(s):
                    return line
        return None

    def in_method_match(
        self,
        expr: str,
        start: str
    ) -> Optional[Set]:
        """
        Try to match all inside a tcl method.
        """
        v = set()
        for line in self.data:
            if line.startswith(start):
                m = re.findall(expr, line)
                if m is False or m is None:
                    return None
                v.update(m)
        return v

    @staticmethod
    def get_sourced_files(
        file: str,
        include_self: bool = True
    ) -> Tuple[List['tcl'], List[str]]:
        """
        Recursively get a list of sourced files.
        Returns a list of tcl objects
        """
        file = path.abspath(file)
        files = [file] if include_self else []
        tcls = []

        def parse(file_):
            if not path.isfile(file_):
                logger.warning(f"{file_}: File doesn't exist!")
                return
            tcl_ = tcl(file_)
            tcls.append(tcl_)
            skip_list = ["adi_env.tcl", "adi_project_xilinx.tcl", "adi_project_intel.tcl", "adi_board.tcl"]

            for i, line in enumerate(tcl_):
                if line.startswith("source "):
                    item = line.split()[1]
                    if any([item.endswith(i) for i in skip_list]):
                       continue
                    if item.startswith("$ad_hdl_dir/"):
                        item = path.abspath(item[12:])
                    elif tcl_.data[i-1].startswith("if [info exists ad_project_dir"):
                        # Skip the alternative path
                        continue
                    else:
                        item = path.abspath(path.join(path.dirname(file_), item))
                    if item not in files:
                        files.append(item)
                        parse(item)
        parse(file)

        # Convert to relative to file dir
        dir_ = path.dirname(file)
        files = [path.relpath(f, dir_)  for f in files]

        return tcls, files
